// =================================================================================
//	IZColor						(c) 2018 Mark F. Coniglio -- All Rights Reserved
// =================================================================================

#ifndef _H_IZColor
#define _H_IZColor

#include "IZQDTypes.h"

struct IzHSVColor {
	float			hue;					// 0.0 to 1.0 represents 0.0 to 360.0
	float			saturation;				// 0.0 to 1.0
	float			value;					// 0.0 to 1.0
	float			alpha;					// 0.0 to 1.0
};

struct IzHSLColor {
	float			hue;					// 0.0 to 1.0 represents 0.0 to 360.0
	float			saturation;				// 0.0 to 1.0
	float			lightness;				// 0.0 = black, 0.5 = fully saturated, 1.0 = white
	float			alpha;					// 0.0 to 1.0
};


// --------------------------------------------------
// IZCOLOR - BASIC TYPE

// four 8 bit components representing ARGB with
// the alpha channel in the most significant byte

typedef	uint32_t	IzColor;

// --------------------------------------------------
// IZCOLOR - CONVERSION FUNCTIONS

#ifdef __cplusplus

const IzColor kIzColorTransparent	= 0x00000000;
const IzColor kIzColorBlack			= 0xFF000000;
const IzColor kIzColorWhite			= 0xFFFFFFFF;

inline IzColor
IZ_ColorMake(
	uint8_t		a,		// 0 to 255
	uint8_t		r,		// 0 to 255
	uint8_t		g,		// 0 to 255
	uint8_t		b)		// 0 to 255
{
	const IzColor aa = ((uint32_t) a << 24);
	const IzColor rr = ((uint32_t) r << 16) & 0x00FF0000;
	const IzColor gg = ((uint32_t) g << 8) & 0x0000FF00;
	const IzColor bb = ((uint32_t) b << 0) & 0x000000FF;

	return aa | rr | gg | bb;
}

// convert a Mac Style QuickTime to an IzColor
inline IzColor
RGBColorToIzColor(
	const RGBColor&		inRGB)
{
	return (
		0xFF000000
		| ((((UInt32)(inRGB.red)) << 8) & 0x00FF0000)
		| ((((UInt32)(inRGB.green)) << 0) & 0x0000FF00)
		| ((((UInt32)(inRGB.blue)) >> 8) & 0x000000FF)
	);
}

// convert and IzColor to a QuickTime style RGBColor
inline RGBColor
IzColorToRGBColor(
	const IzColor		inColor)
{
	RGBColor rgb = {
		static_cast<unsigned short>(((((inColor) >> 8) & 0x0000FF00) | (((inColor) >> 16) & 0x000000FF))),
		static_cast<unsigned short>(((((inColor) >> 0) & 0x0000FF00) | (((inColor) >> 8) & 0x000000FF))),
		static_cast<unsigned short>(((((inColor) << 8) & 0x0000FF00) | (((inColor) >> 0) & 0x000000FF)))
	};
	
	return rgb;
}

// --------------------------------------------------
// MAC (OBJECTIVE-C) SPECIFIC CONVERSIONS

#ifdef _OBJC_OBJC_H_

NSColor*
IzColorToNSColor(
	IzColor			inColor);

IzColor
NSColorToIzColor(
	NSColor*		inColor);


#endif	// #ifdef _OBJC_OBJC_H_

#if TARGET_OS_MAC
// convert an IzColor to a RGBA CGFloat[4] array
inline void
IzColorToRGBACGFloatArray(
	IzColor			inColor,
	CGFloat*		outARGB)
{
	outARGB[3] = ((inColor >> 24) & 0xFF) / 255.0;
	outARGB[0] = ((inColor >> 16) & 0xFF) / 255.0;
	outARGB[1] = ((inColor >> 8) & 0xFF) / 255.0;
	outARGB[2] = ((inColor >> 0) & 0xFF) / 255.0;
}
#endif

// --------------------------------------------------
// WINDOWS SPECIFIC CONVERSIONS

#if TARGET_OS_WIN

// convert a Windows COLORREF to an IzColor
inline IzColor
COLORREFToIzColor(
	const COLORREF		inRGB)
{
	DWORD a = 255;
	DWORD r = GetRValue(inRGB);
	DWORD g = GetGValue(inRGB);
	DWORD b = GetBValue(inRGB);

	return ((((IzColor) ((UInt8) (a))) << 24) | (((IzColor) ((UInt8) (r))) << 16) | (((IzColor) ((UInt8) (g))) << 8) | (((IzColor) (b))));
}

// convert an IzColor to a Windows COLORREF
inline COLORREF
IzColorToCOLORREF(
	const IzColor		inColor)
{
	// UInt32 a = (inColor >> 24) & 0xFF;
	UInt32 r = (inColor >> 16) & 0xFF;
	UInt32 g = (inColor >> 8) & 0xFF;
	UInt32 b = (inColor >> 0) & 0xFF;

	return RGB(r, g, b);
}

#endif	// #if TARGET_OS_WIN

// convert an IzColor to a ARGB float[4] array
inline void
IzColorToARGBFloatArray(
	IzColor			inColor,
	float*			outARGB)
{
	outARGB[0] = ((inColor >> 24) & 0xFF) / 255.0f;
	outARGB[1] = ((inColor >> 16) & 0xFF) / 255.0f;
	outARGB[2] = ((inColor >> 8) & 0xFF) / 255.0f;
	outARGB[3] = ((inColor >> 0) & 0xFF) / 255.0f;
}

// convert an ARGB float[4] array to an IzColor
inline IzColor
ARGBFloatArrayToIzColor(
	const float*	inARGB)
{
	UInt32 a = inARGB[0] * 255.0f;
	UInt32 r = inARGB[1] * 255.0f;
	UInt32 g = inARGB[2] * 255.0f;
	UInt32 b = inARGB[3] * 255.0f;
	
	UInt32 argb = (a << 24) | (r << 16) | (g << 8) | b;
	return argb;
}

// convert an IzColor to a OpenGL compatible RGBA float[4] array
inline void
IzColorToRGBAFloatArray(
	IzColor			inColor,
	float*			outARGB)
{
	outARGB[3] = ((inColor >> 24) & 0xFF) / 255.0f;
	outARGB[0] = ((inColor >> 16) & 0xFF) / 255.0f;
	outARGB[1] = ((inColor >> 8) & 0xFF) / 255.0f;
	outARGB[2] = ((inColor >> 0) & 0xFF) / 255.0f;
}

// --------------------------------------------------
// HSL/HSV <-> IzColor Conversions

IzColor
IZ_ColorHSLtoRGB(
	const IzHSLColor&	inHSL);

IzHSLColor
IZ_ColorRGBtoHSL(
	const IzColor&		inColor);

IzColor
IZ_ColorHSVtoRGB(
	const IzHSVColor&	inHSL);

IzHSVColor
IZ_ColorRGBtoHSV(
	const IzColor&		inColor);

void
QDHSL2RGB(
  const QDHSLColor*	hColor,
  RGBColor*			rColor);

void 
QDRGB2HSL(
  const RGBColor*	rColor,
  QDHSLColor*		hColor);

void
QDHSV2RGB(
  const QDHSVColor*	hColor,
  RGBColor*			rColor);

void 
QDRGB2HSV(
  const RGBColor*	rColor,
  QDHSVColor*		hColor);

// --------------------------------------------------
// USER INTERFACE COLORS

// THEME COLORS
enum IzThemeColorID
{
	kThemeColor_Background,
	kThemeColor_UnfocusedColor,
	kThemeColor_FocusedColor,
	kThemeColor_HighlightColor,
	kThemeColor_GroupBoxBorder,
	kThemeColor_DialogBackground,
	
	kThemeColor_StaticTextEnabled,
	kThemeColor_StaticTextDisabled,
	kThemeColor_TextFieldTextEnabled,
	kThemeColor_TextFieldTextDisabled,
	kThemeColor_TextFieldBkgEnabled,
	kThemeColor_TextFieldBkgDisabled,

	kThemeColor_MaxColors,
	
	kThemeColor_Invalid = -1
};

IzColor		IZ_ColorGetThemeColor(
				IzThemeColorID		inTheme);

void		IZ_ColorSetThemeColor(
				IzThemeColorID		inTheme,
				IzColor				inColor);

// HIGHLIGHT COLOR

// get the text highlight color for the User Interface
IzColor		IZ_ColorGetHighlightColor();

// get the "focus ring" highlight color dor the User Interface
IzColor		IZ_ColorFocusRingColor();

// make an IzColor color darker by 50%
inline void
IZ_DimTextColor(
	RGBColor&		ioColor)
{
	ioColor.red   = (UInt16) ((ioColor.red   + 65535) >> 1);
	ioColor.green = (UInt16) ((ioColor.green + 65535) >> 1);
	ioColor.blue  = (UInt16) ((ioColor.blue  + 65535) >> 1);
}

#endif	// #ifdef __cplusplus

#endif	// #ifndef _H_IZAlert
