// =================================================================================
//	IZSystem						(c) 2018 Mark F. Coniglio -- All Rights Reserved
// =================================================================================

#ifndef _H_IZSystem
#define _H_IZSystem

#include "IZDefinitions.h"
#include "IZTime.h"

#ifdef __cplusplus

#include <string>

// ---------------------------------------------------------------------------
// NSFormattedText - pasteboard item class
// ---------------------------------------------------------------------------
// 3.0.9b15

#if defined(_OBJC_OBJC_H_) && defined(__cplusplus)

	static NSString* NSPasteboardTypeFormattedText = @"com.troikatronix.formatted-text";
	
	@interface NSFormattedText : NSObject <NSCopying, NSPasteboardWriting, NSPasteboardReading, NSCoding>
	{
		Handle	mFormattedTextData;
	}
	- (NSFormattedText*)initWithFormattedText:(Handle)inFormattedText;
	- (Handle) getFormattedTextData;
	@end
	typedef NSFormattedText					IzFormattedText;

#else

	typedef struct OpaqueNSFormattedText*	IzFormattedText;

#endif

// ---------------------------------------------------------------------------
//	IzCondition
// ---------------------------------------------------------------------------
// A condition that becomes true when the specified block of memory
// becomes non-zero.

class IzCondition
{
public:
					IzCondition();
					
					IzCondition(
						const void*		inConditionPtr,
						const size_t	inConditionLength);
			
					~IzCondition();
	
	virtual bool	IsConditionSet() const;

protected:
	const void*	mConditionPointer;
	size_t		mConditionLength;
	void*		mZeroCond;
};

// ---------------------------------------------------------------------------
//	FOLDER PATHS
// ---------------------------------------------------------------------------

typedef OSType	IzSysFolderCode;

enum {

	kIzSysFolderCode_Applications		= FOUR_CHAR_CODE('appl'),
	kIzSysFolderCode_Documents			= FOUR_CHAR_CODE('docs'),
	kIzSysFolderCode_Desktop			= FOUR_CHAR_CODE('desk'),
	
	#if TARGET_OS_MAC
	kIzSysFolderCode_SystemLibrary		= FOUR_CHAR_CODE('slib'),
	kIzSysFolderCode_Library			= FOUR_CHAR_CODE('libr'),
	kIzSysFolderCode_UserLibrary		= FOUR_CHAR_CODE('uibr'),
	#endif
	
	// kIzSysFolderCode_Preferences:
	// MacOS: ~/Library/Preferences
	// Win32: X:\Users\USERNAME\AppData\Roaming\TroikaTronix
	kIzSysFolderCode_Preferences		= FOUR_CHAR_CODE('pref'),
	
	// kIzSysFolderCode_ApplicationData:
	// MacOS:    /Library/Application Support
	// Win32 32: C:\Program Files (x86)\Common Files
	// Win32 64: C:\Program Files\Common Files
	kIzSysFolderCode_ApplicationData	= FOUR_CHAR_CODE('asup'),

	// kIzSysFolderCode_SharedData:
	// MacOS: /Users/Shared
	// Win32: X:\ProgramData
	kIzSysFolderCode_SharedData			= FOUR_CHAR_CODE('sdat'),

	// kIzSysFolderCode_TemporaryItems:
	// MacOS: /tmp or ??
	// Win32: X:\Windows\Temp
	kIzSysFolderCode_TemporaryItems		= FOUR_CHAR_CODE('temp'),

	// kIzSysFolderCode_LowLevelFolder:
	// MacOS: /usr/local/troikatronix
	// Win32: ???
	kIzSysFolderCode_LowLevelFolder		= FOUR_CHAR_CODE('tllf'),
	
	// Same result as kIzSysFolderCode_ApplicationData
	kIzSysFolderCode_GlobalAudioPlugins	= FOUR_CHAR_CODE('gaup'),
	
	// MacOS: Same result as kIzSysFolderCode_UserLibrary
	// Win32: Same result as kIzSysFolderCode_ApplicationData
	kIzSysFolderCode_UserAudioPlugins	= FOUR_CHAR_CODE('uaup')
	
};

std::string		IZ_SystemGetPathToSystemFolder(
					OSType		inIzSysFolderCode,
					bool		inWithTrailingDelimiter = false);

std::string		IZ_SystemFileUrlToNativePath(
					std::string		inFileURL);

std::string		IZ_SystemNativePathToFileUrl(
					std::string		inFileURL);

// ---------------------------------------------------------------------------
//	ERROR LOGGING
// ---------------------------------------------------------------------------
// added 3.1.2b11

void
IZ_SystemSetLogFileName(
	std::string			inFileNameAndExtension);
	
void
IZ_SystemWriteErrorToLogFile(
	std::string			inErrorType,
	std::string			inErrorMesasge);
	

// ---------------------------------------------------------------------------
//	MISCELLEANEOUS FUNCTIONS
// ---------------------------------------------------------------------------
size_t			IZ_SystemGetInstalledPhysicalRAMBytes();

IzMilliseconds	IZ_SystemGetDoubleClickIntervalMilliseconds();

bool			IZ_SystemPutTextOnClipboard(
					std::string		inString,
					bool			inClearClipcboard = true);	// inClearClipcboard added for 3.0.9b15

std::string		IZ_SystemGenerateUUID();

#if TARGET_OS_MAC
void			IZ_OpenMissionControlSystemPreference();
#endif

#if TARGET_OS_WIN
UINT			IZ_SystemGetRTFClipboardFormat();			// added for 3.0.9b15
UINT			IZ_SystemGetFormattedTextClipboardFormat();	// added for 3.0.9b15
#endif

// ------------------------------------------------------------
// WINDOWS ONLY - LINE ENDING HELPERS
// ------------------------------------------------------------

enum IzLineEnding
{
	kIzLineEnding_CR = 0,
	kIzLineEnding_LF,
	kIzLineEnding_CRLF
};

#if TARGET_OS_MAC
	#define kIzLineEnding_SystemNative		kIzLineEnding_CR
	#define kIzLineEnding_TextFieldNative	kIzLineEnding_CR
	#define kIzLineEnding_ScrapNative		kIzLineEnding_CR
#elif TARGET_OS_WIN
	#define kIzLineEnding_SystemNative		kIzLineEnding_CRLF
	#define kIzLineEnding_TextFieldNative	kIzLineEnding_CR
	#define kIzLineEnding_ScrapNative		kIzLineEnding_CRLF
#endif

size_t
IZ_AllocUnicodeTextEnsuringLineEnding(
	IzLineEnding		inLineEnding,
	const IzUniChar*	inText,
	size_t				inTextCharLen,
	IzUniChar**			outText);

size_t
IZ_AllocUTF8TextEnsuringLineEnding(
	IzLineEnding		inLineEnding,
	const char*			inText,
	size_t				inTextCharLen,
	char**				outText);


// ---------------------------------------------------------------------------
// SCRAP RELATED CALLS + TYPDEFS
// ---------------------------------------------------------------------------

typedef OSType   IzScrapFlavorType;

enum {
	kIzScrapFlavorTypePicture			= FOUR_CHAR_CODE('PICT'), // CGBitmapContext on MacOS, HBITMAP on Windows
	kIzScrapFlavorTypeText				= FOUR_CHAR_CODE('TEXT'), // stream of 8 bit characters
	kIzScrapFlavorTypeStyledText		= FOUR_CHAR_CODE('styl'), // RTF text on both platforms (3.0.9b15))
	kIzScrapFlavorFormattedText			= FOUR_CHAR_CODE('IZFT'), // Isadora CFormattedText as a handle (3.0.9b15)
	// kIzScrapFlavorTypeMovie			= FOUR_CHAR_CODE('moov'), // currently not used
	kIzScrapFlavorTypeSound				= FOUR_CHAR_CODE('snd '), // AIFF on MacOS, WAVE on Windows
	kIzScrapFlavorTypeUnicode			= FOUR_CHAR_CODE('utxt'), // stream of UTF16 characters
	// kIzScrapFlavorTypeStyledUnicode	= FOUR_CHAR_CODE('ustl'), // currently unused

	// if you wish to use your own scrap flavors
	// then must be within this range of values
	kIzScrapFlavorUserFirst				= 0x200,
	kIzScrapFlavorUserLast				= 0x2FF
};


OSStatus	IzPutScrap(
				IzScrapFlavorType	inFlavorType,
				size_t				inSourceBufferByteCount,
				const void*			inSourceBuffer,
				bool				inClearClipboard = true);

OSStatus	IzHasScrap(
				IzScrapFlavorType	inFlavorType);

OSStatus	IzGetScrap(
				Handle				inDestination,
				IzScrapFlavorType	inFlavorType,
				size_t*				outLength,
				size_t*				outOffset);

OSStatus	IzZeroScrap();

#endif	// #ifdef __cplusplus

#endif	// #ifndef _H_IZSystem
