// ===========================================================================
//	AnalyzeImageDemo			(c) 2021 Mark F. Coniglio. All rights reserved
// ===========================================================================

/*

# ANALYZE IMAGE DEMO PLUGIN - READ ME

## DESCRIPTION

(Please don't skip the section below "ABOUT ANALYZING OPENGL TEXTURE DATA"
as this information is very important!)

This plugin does a simple analysis of the color of the incoming video stream
to illustrate an important general technique: how to use C++ code to analyze
a video frame that typically resides on the graphics card (GPU) of your
computer. This technique is required if you wish to analyze a video stream
with a CPU based analysis library such as OpenCV.

## CUSTOM USER INTERFACE DEMONSTRATION

Instead of an icon, some actors offer a custom, dynamic display that gives
the user an indication of their function. This plugin illustrates one such
interface element by displaying an indicator that shows the portion of the
image that is being analyzed. See the GetActorDefinedArea and
DrawActorDefinedArea functions for more information.

## ABOUT VIDEO TYPES

In Isadora, video data comes in three "flavors":

- **vid-gpu**: Image is stored as an OpenGL texture on your graphics card.
Image data is processed using OpenGL shaders

- **vid-ci**: Image is stored as
a CIImage on your graphics card (macOS only). Image data is processed using
OpenGL shaders

- **vid-cpu**: Image is stored in the main memory of your
computer. image data is processed using standard C++.  format is only used by
legacy Isadora plugins, usually marked by the word classic" in their name.

This plugin offers a mutable input that will "mutate" to whatever video type
is received at the input. Then a conversion function supplied by the SDK
ensures that, regardless of the video type you receive, the image ends
up as a bitmap in the main memory of the computer that can be analyzed using
C++.

## **VERY IMPORTANT**: ABOUT ANALYZING OPENGL TEXTURE DATA

The main processor (CPU) cannot directly read the individual pixels of a
texture that resides within the video memory of the GPU. To gain access to
those pixels, one must transfer the image from the GPU's video memory to the
computer's main memory (RAM) so that the computer's processor can access it.

This plugin illustrates how to use a function offered by the SDK called
`ConvertVideoStreamToBitmap_()` to pull texture data back into the computer's
main memory. This function is designed to be as efficient as possible, but
pulling texture data back to the the main processor's memory is,
unfortunately, inherently inefficient.

Why? Because doing so causes a "stall", where all pending parallel operations
pending on the GPU must be completed before the pixels can be read back to
the CPU. This can significantly reduce the throughput of the GPU, which will
in turn slow down all video processing in the currently executing Isadora
Scene.

If this is done by many actors in the same processing cycle, Isadora's
LOAD indicator may go up significantly as all video processing is slowed
down because of multiple stalls. This may be especially noticeable this
is done with higher resolution images such as 2K or (worse yet) 4K video.
That is why this plugin offers the opportunity to sample only a portion of
the image, because pulling large images can introduce even more significant
delays.

To always analyze the entire image, set the `ANALAYZE_PARTIAL_IMAGE` switch
to 0. This will remove all code that has to do with selecting a portion of
the image for analysis.

## CUSTOMIZING THIS SOURCE CODE

To customize this file, search for the text ###. All of the places where you
will need to customize the file are marked with this pattern of characters.

## LICENSE

IMPORTANT: This source code ("the software") is supplied to you in
consideration of your agreement to the following terms. If you do not agree
to the terms, do not install, use, modify or redistribute the software.

Mark Coniglio (aka TroikaTronix) grants you a personal, non exclusive license
to use, reproduce, modify this software with and to redistribute it, with or
without modifications, in source and/or binary form. Except as expressly
stated in this license, no other rights are granted, express or implied, to
you by TroikaTronix.

This software is provided on an "AS IS" basis. TroikaTronix makes no
warranties, express or implied, including without limitation the implied
warranties of non-infringement, merchantability, and fitness for a
particular purpurse, regarding this software or its use and operation alone
or in combination with your products.

In no event shall TroikaTronix be liable for any special, indirect,
incidental, or consequential damages arising in any way out of the use,
reproduction, modification and/or distribution of this software.

*/

// ---------------------------------------------------------------------------------
// INCLUDES
// ---------------------------------------------------------------------------------

// ISADORA INCLUDES
#include "IsadoraPluginPrefix.h"
#include "IsadoraCallbacks.h"
#include "PluginDrawUtil.h"

// STANDARD INCLUDES
#include <string.h>
#include <stdio.h>
#include <math.h>

// ---------------------------------------------------------------------------------
// MacOS Specific
// ---------------------------------------------------------------------------------
#if TARGET_OS_MAC
#define EXPORT_
#endif

// ---------------------------------------------------------------------------------
// Win32  Specific
// ---------------------------------------------------------------------------------
#if TARGET_OS_WIN	
	#define EXPORT_ __declspec(dllexport)
#endif

// ---------------------------------------------------------------------------------
//	Exported Function Definitions
// ---------------------------------------------------------------------------------

#ifdef __cplusplus
extern "C" {
#endif

EXPORT_ void GetActorInfo(void* inParam, ActorInfo* outActorParams);

#ifdef __cplusplus
}
#endif

// ---------------------------------------------------------------------------------
//	FORWARD DECLARTIONS
// ---------------------------------------------------------------------------------
static void
ReceiveMessage(
	IsadoraParameters*	ip,
	MessageMask			inMessageMask,
	PortIndex			inPortIndex1,
	const MsgData*		inData,
	UInt32				inLen,
	IzRefCon			inRefCon);

// ---------------------------------------------------------------------------------
// CONSTANTS
// ---------------------------------------------------------------------------------
// If you know you will always analyze the full image, set this switch to 0. It will
// then ignore all code related to analyzing a partial image, including the custom
// user interface.

#define ANALAYZE_PARTIAL_IMAGE				0

// ---------------------------------------------------------------------------------
// CONSTANTS
// ---------------------------------------------------------------------------------
#if ANALAYZE_PARTIAL_IMAGE
const SInt16	kIndicatorEnclSize			= 32;
const SInt16	kIndicatorInset				= 1;
const SInt16	kIndicatorSize				= (kIndicatorEnclSize - kIndicatorInset*2);
#endif

// ---------------------------------------------------------------------------------
// GLOBAL VARIABLES
// ---------------------------------------------------------------------------------
// ### Declare global variables, common to all instantiations of this plugin here
//
// Note that global variables may be reinitialized if the plugin (dylib on macOS, DLL
// on Windows) is entirely disposed by Isadora.
//
// For example, if one loads an Isadora document with this actor and subequently
// closes the file, there will no longer beany activce refefernces to the plugin
// and it will be deallocated.
//
// If you then open the same file the plugin, it's global variables, will be
// initialized again.

// Example: static int gMyGlobalVariable = 5;

// ---------------------------------------------------------------------------------
// PluginInfo struct
// ---------------------------------------------------------------------------------
// ### This structure neeeds to contain all variables used by your plugin. Memory for
// this struct is allocated during the CreateActor function, and disposed during
// the DisposeActor function, and is private to each copy of the plugin.
//
// If your plugin needs global data, declare them as static variables within this
// file. Any static variable will be global to all instantiations of the plugin.

typedef struct {

	ActorInfo*				mActorInfoPtr;		// our ActorInfo Pointer - set during create actor function
	MessageReceiverRef		mMessageReceiver;	// pointer to our message receiver reference
	
	bool					mNeedsRender;		// ### set to true when we receive a frame

#if ANALAYZE_PARTIAL_IMAGE
	bool					mNeedsReposition;	// ### set to true when the position or size changes

	IzRect					mSrcRect;			// ### rectangle of color we are analying
	float					mSrcWidth;			// ### source width from 0 to 1 (i.e., 0 to 100%)
	float					mSrcHeight;			// ### source height from 0 to 1 (i.e., 0 to 100%)
	float					mSrcLeft;			// ### left edge of the source from 0 to 1 (i.e., 0 to 100%)
	float					mSrcTop;			// ### top edge of the source from 0 to 1 (i.e., 0 to 100%)

	IzRect					mIndicatorRect;		// ### indicator rect in the Actor Draw Area (user interface)
#endif

	Value					mInputBuffer;		// ### the most recent input value
	
	IzzyBitmapConverter		mBitmapConverter;	// ### the converter used to pull GPU video back into the CPU
	
} PluginInfo;

// A handy macro for casting the mActorDataPtr to PluginInfo*
#if __cplusplus
#define	GetPluginInfo_(actorDataPtr)		static_cast<PluginInfo*>((actorDataPtr)->mActorDataPtr);
#else
#define	GetPluginInfo_(actorDataPtr)		(PluginInfo*)((actorDataPtr)->mActorDataPtr);
#endif

// ---------------------------------------------------------------------------------
//	Constants
// ---------------------------------------------------------------------------------
//	Defines various constants used throughout the plugin.

// CONSTANTS FOR GPU VERSION

// ### GROUP ID
// Define the group under which this plugin will be displayed in the Isadora interface.
// These are defined under "Actor Types" in IsadoraTypes.h

static const OSType	kActorClass 	= kGroupVideo;

// ### PLUGIN IN
// Define the plugin's unique four character identifier. Contact TroikaTronix to
// obtain a unique four character code if you want to ensure that someone else
// has not developed a plugin with the same code. Note that TroikaTronix reserves
// all plugin codes that begin with an underline, an at-sign, and a pound sign
// (e.g., '_', '@', and '#'.)

static const OSType	kActorID		= FOUR_CHAR_CODE('ABDM');

// ### ACTOR NAME
// The name of the actor. This is the name that will be shown in the User Interface.

static const char* kActorName		= "Analyze Image Demo";

// ### PROPERTY DEFINITION STRING
// The property string. This string determines the inputs and outputs for your plugin.
// See the IsadoraCallbacks.h under the heading "PROPERTY DEFINITION STRING" for the
// meaning ofthese codes. (The IsadoraCallbacks.h header can be seen by opening up
// the IzzySDK Framework while in the Files view.)
//
// IMPORTANT: You cannot use spaces in the property name. Instead, use underscores (_)
// where you want to have a space.
//
// Note that each line ends with a carriage return (\r), and that only the last line of
// the bunch ends with a semicolon. This means that what you see below is one long
// null-terminated c-string, with the individual lines separated by carriage returns.

static const char* sPropertyDefinitionString =

		// INPUT PROPERTY DEFINITIONS
	"INPROP video_in		vin		data		texture,mutable1,noconform	* * 0	* * 0	0	texture,video,ci-image\r"
#if ANALAYZE_PARTIAL_IMAGE
	"INPROP width			wid		float		number				0		100		100\r"
	"INPROP height			heig	float		number				0		100		100\r"
	"INPROP horiz.			horz	float		number				0		100		0\r"
	"INPROP vert.			vert	float		number				0		100		0\r"
#endif

		// OUTPUT PROPERTY DEFINITIONS
	"OUTPROP red			cred	int			number				0		255		0\r"
	"OUTPROP green			cgrn	int			number				0		255		0\r"
	"OUTPROP blue			cblu	int			number				0		255		0\r"
	"OUTPROP alpha			calp	int			number				0		255		0\r"
	"OUTPROP color			colr	int			acolor,hex			*		*		4294901760\r";

// ### Property Index Constants
// Properties are referenced by a one-based index. The first input property will
// be 1, the second 2, etc. Similarly, the first output property starts at 1.
// You whould have one constant for each input and output property defined in the 
// property definition string.

enum
{
// INPUTS
	kInputVideo1 = 1,
#if ANALAYZE_PARTIAL_IMAGE
	kInputWidth,
	kInputHeight,
	kInputHoriz,
	kInputVert,
#endif

// OUTPUTS
	kOutputRed = 1,
	kOutputGreen,
	kOutputBlue,
	kOutputAlpha,
	kOutputColor
};


// ---------------------
//	Help String
// ---------------------
// ### Help Strings
//
// The first help string is for the actor in general. This followed by help strings
// for all of the inputs, and then by the help strings for all of the outputs. These
// should be given in the order that they are defined in the Property Definition
// String above.
//
// In all, the total number of help strings should be (num inputs + num outputs + 1)
//
// Note that each string is followed by a comma -- it is a common mistake to forget the
// comma which results in the two strings being concatenated into one.

const char* sHelpStrings[] =
{
	// ACTOR HELP
	"Measures the red, green and blue intensity of the incoming video stream. Change the 'width', "
		"'height', 'horiz.' and 'vert.' inputs to specify a specific area to be measured.",

	// INPUT HELP
	"The video stream whose red, green, blue and alpha components will be measured.",
	
#if ANALAYZE_PARTIAL_IMAGE
	"The width of the area to be measured, from 0 to 100% of the source video stream's width. (Note that "
		"setting this value to 0 will measure only one pixel.)",
	
	"The height of the area to be measured, from 0 to 100% of the source video stream's height. (Note that "
		"setting this value to 0 will measure only one pixel.)",
	
	"The horizontal position of the area to be measured. A value of 0% places the left edge of the sampling "
		"area along the left edge of the video image; a value of 100% places the right edge of the "
		"sampling area along the right edge of the video image.",
	
	"The vertical position of the area to be measured. A value of 0% places the top edge of the sampling "
		"area along the top edge of the video image; a value of 100% places the bottom edge of the "
		"sampling area along the bottom edge of the video image.",
#endif

	// OUTPUT HELP
	"The average intensity of the red component of the measured area, from 0 to 255",
	"The average intensity of the green component of the measured area, from 0 to 255",
	"The average intensity of the blue component of the measured area, from 0 to 255",
	"The average intensity of the alpha component of the measured area, from 0 to 255",
	"The average color of the measured area."
};

// ---------------------------------------------------------------------------------
//		CreateActor
// ---------------------------------------------------------------------------------
// Called once, prior to the first activation of an actor in its Scene. The
// corresponding DisposeActor actor function will not be called until the file
// owning this actor is closed, or the actor is destroyed as a result of being
// cut or deleted.

static void
CreateActor(
	IsadoraParameters*	ip,	
	ActorInfo*			ioActorInfo)		// pointer to this actor's ActorInfo struct - unique to each instance of an actor
{
	// creat the PluginInfo struct - initializing it to all zeroes
	PluginInfo* info = (PluginInfo*) IzzyMallocClear_(ip, sizeof(PluginInfo));
	PluginAssert_(ip, info != NULL);
	
	ioActorInfo->mActorDataPtr = info;
	info->mActorInfoPtr = ioActorInfo;

	// ### allocation and initialization of private member variables
	
}

// ---------------------------------------------------------------------------------
//		DisposeActor
// ---------------------------------------------------------------------------------
// Called when the file owning this actor is closed, or when the actor is destroyed
// as a result of its being cut or deleted.
//
static void
DisposeActor(
	IsadoraParameters*	ip,
	ActorInfo*			ioActorInfo)		// pointer to this actor's ActorInfo struct - unique to each instance of an actor
{
	PluginInfo* info = GetPluginInfo_(ioActorInfo);
	PluginAssert_(ip, info != NULL);
	
	// ### destruction of private member variables
	
	// destroy the PluginInfo struct allocated with IzzyMallocClear_ the CreateActor function
	PluginAssert_(ip, ioActorInfo->mActorDataPtr != NULL);
	IzzyFree_(ip, ioActorInfo->mActorDataPtr);
}

// ---------------------------------------------------------------------------------
//		DisposeVideoConverter
// ---------------------------------------------------------------------------------
// Dispose the video converter, which converts texture or core image video inputs
// to a CPU based bitmaps.
static void
DisposeVideoConverter(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo)
{
	PluginInfo*	info = (PluginInfo*) inActorInfo->mActorDataPtr;

	// if the bitmap convert exists, then dispose it
	if (info->mBitmapConverter != NULL) {
		DisposeIzzyBitmapConverter_(ip, inActorInfo, info->mBitmapConverter);
		info->mBitmapConverter = NULL;
	}
	
	// we need to generate an output as soon as we receive more input
	info->mNeedsRender = true;
}

// ---------------------------------------------------------------------------------
//		ActivateActor
// ---------------------------------------------------------------------------------
//	Called when the scene that owns this actor is activated or deactivated. The
//	inActivate flag will be true when the scene is activated, false when deactivated.
//
static void
ActivateActor(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,		// pointer to this actor's ActorInfo struct - unique to each instance of an actor
	Boolean				inActivate)			// true when actor is becoming active, false otherwise.
{
	PluginInfo* info = GetPluginInfo_(inActorInfo);
	
	// ------------------------
	// ACTIVATE
	// ------------------------
	
	if (inActivate) {
	
		// Isadora passes various messages to plugins that request them.
		// These include Mouse Moved messages, Key Down/Key Up messages,
		// Video Frame Clock messages, etc. The complete list can be found
		// in the enumeration in MessageReceiverCommon.h
		
		// You ask Isadora for these messages by calling CreateMessageReceiver_
		// with a pointer to your function, and the message types you would
		// like to receive. (These are bitmapped flags, so you can combine as
		// many as you like: kWantKeyDown | kWantKeyDown for instance.)
		
		// Here we request that our ReceiveMessage function is called
		// whenever the Isadora New Video Frame message is sent,
		// which happens periodically, 30 times per second. We set the ref
		// con to our ActorInfo ptr so that we can access that information
		// from ReceiveMessage callback.
		
		// we should not already have a message receiver
		PluginAssert_(ip, info->mMessageReceiver == NULL);
		
		// create a message receiver that will be notified of
		// video frame ticks
		info->mMessageReceiver = CreateActorMessageReceiver_(
			ip,
			inActorInfo,
			ReceiveMessage,
			0,
			kWantVideoFrameTick | kWantStageAboutToDispose,		// GPU: add kWantStageAboutToDispose
			PtrToIzRefCon(inActorInfo));

	
	// ------------------------
	// DEACTIVATE
	// ------------------------
	
	} else {
	
		// ensure the video converter is disposed
		DisposeVideoConverter(ip, inActorInfo);
		
		// dispose our message receiver when we are deactivated.
		if (info->mMessageReceiver != NULL) {
			DisposeMessageReceiver_(ip, info->mMessageReceiver);
			info->mMessageReceiver = NULL;
		}
		
		// ### dispose any data that you don't need when
		// you are not active.
	}
}

// ---------------------------------------------------------------------------------
//		GetParameterString
// ---------------------------------------------------------------------------------
//	Returns the property definition string. Called when an instance of the actor
//	needs to be instantiated.

static const char*
GetParameterString(
	IsadoraParameters*	/* ip */,
	ActorInfo*			/* inActorInfo */)
{
	return sPropertyDefinitionString;
}

// ---------------------------------------------------------------------------------
//		GetHelpString
// ---------------------------------------------------------------------------------
//	Returns the help string for a particular property. If you have a fixed number of
//	input and output properties, it is best to use the PropertyTypeAndIndexToHelpIndex_
//	function to determine the correct help string to return.

static void
GetHelpString(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	PropertyType		inPropertyType,			// kPropertyTypeInvalid when requesting help for the actor
												// or kInputProperty or kOutputProperty when requesting help for a specific property
	PropertyIndex		inPropertyIndex1,		// the one-based index of the property (when inPropertyType is not kPropertyTypeInvalid)
	char*				outParamaterString,		// receives the help string
	UInt32				inMaxCharacters)		// size of the outParamaterString buffer
{
	const char* helpstr = NULL;
	
	// The PropertyTypeAndIndexToHelpIndex_ converts the inPropertyType and
	// inPropertyIndex1 parameters to determine the zero-based index into
	// your list of help strings.
	UInt32 index1 = PropertyTypeAndIndexToHelpIndex_(ip, inActorInfo, inPropertyType, inPropertyIndex1);
	
	// get the help string
	helpstr = sHelpStrings[index1];
	
	// copy it to the output string
	strncpy(outParamaterString, helpstr, inMaxCharacters);
}
	
// ---------------------------------------------------------------------------------
//		HandlePropertyChangeValue	[INTERRUPT SAFE]
// ---------------------------------------------------------------------------------
//	### This function is called whenever one of the input values of an actor changes.
//	The one-based property index of the input is given by inPropertyIndex1.
//	The new value is given by inNewValue, the previous value by inOldValue.
//
static void
HandlePropertyChangeValue(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	PropertyIndex		inPropertyIndex1,			// the one-based index of the property than changed values
	ValuePtr			/* inOldValue */,			// the property's old value
	ValuePtr			inNewValue,					// the property's new value
	Boolean				inInitializing)				// true if the value is being set when an actor is first initalized
{
	PluginInfo* info = GetPluginInfo_(inActorInfo);

	switch (inPropertyIndex1) {
	case kInputVideo1:
		{
			info->mInputBuffer = *inNewValue;
			info->mNeedsRender = true;
		}
		break;
	
	#if ANALAYZE_PARTIAL_IMAGE
	
	case kInputVert:
		{
			info->mSrcTop = float(inNewValue->u.fvalue / 100.0);
			info->mNeedsReposition = true;
			info->mNeedsRender = true;
		}
		break;
	case kInputHoriz:
		{
			info->mSrcLeft = float(inNewValue->u.fvalue / 100.0);
			info->mNeedsReposition = true;
			info->mNeedsRender = true;
		}
		break;
	case kInputWidth:
		{
			info->mSrcWidth = float(inNewValue->u.fvalue / 100.0);
			info->mNeedsReposition = true;
			info->mNeedsRender = true;
		}
		break;
	case kInputHeight:
		{
			info->mSrcHeight = float(inNewValue->u.fvalue / 100.0);
			info->mNeedsReposition = true;
			info->mNeedsRender = true;
		}
		break;
		
	#endif
	
	}

	#if ANALAYZE_PARTIAL_IMAGE
	
	// ### CUSTOM USER INTERFACE
	// if the source position or size was changed at all, we need
	// to recalculate the indicator rect for the user interface
	
	if (info->mNeedsReposition) {
		
		float gsScaleFactor = GetActorGraphicScaling_(ip, inActorInfo);
		const IzCoord indicatorInset = IzCoord(ActorPixelToScreen(kIndicatorInset));
		const IzCoord indicatorSize = IzCoord(ActorPixelToScreen(kIndicatorSize));

		IzRect area;
		IzzySetRect(&area, indicatorInset, indicatorInset, indicatorInset+indicatorSize, indicatorInset+indicatorSize);

		IzCoord srcWidth = IzCoord(indicatorSize * info->mSrcWidth);
		IzCoord srcHeight = IzCoord(indicatorSize * info->mSrcHeight);
		if (srcWidth < 1)
			srcWidth = 1;
		if (srcHeight < 1)
			srcHeight = 1;
		
		area.left = IzCoord((indicatorSize - srcWidth) * info->mSrcLeft);
		area.top = IzCoord((indicatorSize - srcHeight) * info->mSrcTop);
		area.right = area.left + srcWidth;
		area.bottom = area.top + srcHeight;
		
		if (!IzzyEqualRect(&area, &info->mIndicatorRect)) {
			info->mIndicatorRect = area;
			inActorInfo->mDrawActorAreaFlags = kDrawAllActorAreas;
		}
	}
	#endif

}

// ---------------------------------------------------------------------------------
//		HandlePropertyChangeType
// ---------------------------------------------------------------------------------
// ### This actor allows input from all three of teh video types supported by
// Isadora: vid-gpu (an OpenGL texture living on the graphics card), vid-ci (macOS
// only core image video, which also lives on the graphcisc card), and vid-cpu
// which is a CPU based bitmap.
//
// This function is called when the user connects a link that is streaming a
// different video type. It ensure that all video inputs are updated to the
// matching type.

typedef struct {
	PropertyType	type;
	PropertyIndex	index;
} PropItem;

static PropItem sPropItems[1] = {
	{ kInputProperty, kInputVideo1 }
};

static void
HandlePropertyChangeType(
	IsadoraParameters*					ip,
	ActorInfo*							inActorInfo,
	const PropertyDefinedValueSource*	inPropDefValueSrc,
	PropertyType						inPropertyType,
	PropertyIndex						inPropertyIndex1,
	PropertyDispFormat					inAvailableFormats,
	PropertyDispFormat					inCurFormat,
	ValuePtr							inMinimum,
	ValuePtr							inMaximum,
	ValuePtr							inScaleMin,
	ValuePtr							inScaleMax,
	ValuePtr							inValue)
{
	// static varible prevents recursion which might happen in certain situations
	static bool sHandlePropertyChangeType = false;
	
	if (!sHandlePropertyChangeType) {
	
		sHandlePropertyChangeType = true;
		
		const size_t kMaxVideoProperties = sizeof(sPropItems) / sizeof(PropItem);
		
		for (size_t i=0; i<kMaxVideoProperties; i++) {
			if (sPropItems[i].type == inPropertyType && sPropItems[i].index == inPropertyIndex1) {
				for (size_t j=0; j<kMaxVideoProperties; j++) {
					if (j != i) {
						ChangePropertyType_(ip, inActorInfo, inPropDefValueSrc, sPropItems[j].type, sPropItems[j].index, inAvailableFormats, inCurFormat, inMinimum, inMaximum, inScaleMin, inScaleMax, inValue);
					}
				}
				break;
			}
		}
		sHandlePropertyChangeType = false;
	}
}


// ---------------------------------------------------------------------------------
//		GetActorDefinedArea
// ---------------------------------------------------------------------------------
//	If the mGetActorDefinedAreaProc in the ActorInfo struct points to this function,
//	it indicates to Isadora that the object would like to draw either an icon or else
//	an graphic representation of its function.
//
// ### This actor draws a custom, dynamic image in its Actor Defined Area (ADA) to
// help the user visualize what the actor is doing. There are two possible areas
// into which one can draw: the "top" area which is in the center between the
// input and outputs, and the bottom area which, if defined with a non-zero height,
// is added to the bottom of the actor.
//
// This particular actor usses the top ADA area with a minimum size of 32 x 32 pixels.

static Boolean
GetActorDefinedArea(
	IsadoraParameters*			ip,			
	ActorInfo*					inActorInfo,
	SInt16*						outTopAreaWidth,			// returns the width to reserve for the top Actor Defined Area
	SInt16*						outTopAreaMinHeight,		// returns the minimum height of the top area
	SInt16*						outBotAreaHeight,			// returns the width to reserve for the bottom Actor Defined Area
	SInt16*						outBotAreaMinWidth)			// returns the minimum width of the bottom area
{
	#if ANALAYZE_PARTIAL_IMAGE
		*outTopAreaWidth = 32;
		*outTopAreaMinHeight = 32;
	#else
		*outTopAreaWidth = 0;
		*outTopAreaMinHeight = 0;
	#endif
	
	*outBotAreaHeight = 0;
	*outBotAreaMinWidth = 0;

	return true;
}

// ---------------------------------------------------------------------------------
//		DrawActorDefinedArea
// ---------------------------------------------------------------------------------
//	If GetActorDefinedArea is defined, then this function will be called whenever
//	your ActorDefinedArea needs to be drawn.
//
//	Beacuse we are using the PICT 0 resource stored with this plugin, we can use
//	the DrawActorDefinedAreaPict_ supplied by the Isadora callbacks.
//
//  DrawActorDefinedAreaPict_ is Alpha Channel aware, so you can have nice
//	shading if you like.

static void
DrawActorDefinedArea(
	IsadoraParameters*			ip,
	ActorInfo*					inActorInfo,
	void*						/* inDrawingContext */,		// unused at present
	ActorDefinedAreaPart		inActorDefinedAreaPart,		// the part of the actor that needs to be drawn
	ActorAreaDrawFlagsT			/* inAreaDrawFlags */,		// actor draw flags
	Rect*						inADAArea,					// rect enclosing the entire Actor Defined Area
	Rect*						/* inUpdateArea */,			// subset of inADAArea that needs updating
	Boolean						inSelected)					// TRUE if actor is currently selected
{
	#if ANALAYZE_PARTIAL_IMAGE
	
	PluginInfo* info = GetPluginInfo_(inActorInfo);

	if (inActorDefinedAreaPart == kActorDefinedAreaTop) {
	
		// get a graphics context
		UGraphics* g = GetActorGraphics_(ip, inActorInfo);
		// get the current scale factor which may be something other than 1.0 on windows
		float gsScaleFactor = GetActorGraphicScaling_(ip, inActorInfo);
		// get the indicator enclosing rect scaled by the scale factor
		const IzCoord indicatorEnclSize = IzCoord(ActorPixelToScreen(kIndicatorEnclSize));
		// get the indicator inset scaled by the scale factor
		const IzCoord indicatorInset = IzCoord(ActorPixelToScreen(kIndicatorInset));
		
		ColorRef indicatorColor = RGBToColorRef(0, 179, 0);
		ColorRef bkgColor = RGBToColorRef(0, 102, 0);

		IzRect area = *inADAArea;
		
		if (area.bottom - area.top > indicatorEnclSize)
			area.bottom = area.top + indicatorEnclSize;
		if (area.right - area.left > indicatorEnclSize)
			area.right = area.left + indicatorEnclSize;
		
		// center the background inside the AD Area
		IzzyOffsetRect(&area,
			((inADAArea->right - inADAArea->left) - (area.right - area.left))/2,
			((inADAArea->bottom - inADAArea->top) - (area.bottom - area.top))/2);
		
		// draw a shaded box in green
		DrawShadedBoxPercent_(ip, g, &area, bkgColor, 10, 150, false);

		// draw the indicator in yellow
		IzRect indicator = info->mIndicatorRect;
		IzzyOffsetRect(&indicator, indicatorInset+area.left, indicatorInset+area.top);
		SetForegroundColorRef_(ip, g, indicatorColor);
		FrameRect_(ip, g, &indicator, 1);
	}
	#endif
}
	
// ---------------------------------------------------------------------------------
//		GetActorInfo
// ---------------------------------------------------------------------------------
//	This is function is called by to get the actor's class and ID, and to get
//	pointers to the all of the plugin functions declared locally.
//
//	All members of the ActorInfo struct pointed to by outActorParams have been
//	set to 0 on entry. You only need set functions defined by your plugin
//	
EXPORT_ void
GetActorInfo(
	void*				/* inParam */,
	ActorInfo*			outActorParams)
{
	// REQUIRED information
	outActorParams->mActorName							= kActorName;
	outActorParams->mClass								= kActorClass;
	outActorParams->mID									= kActorID;
	outActorParams->mCompatibleWithVersion				= kActorCompatilibilityVersion_v1;
	
	// REQUIRED functions
	outActorParams->mGetActorParameterStringProc		= GetParameterString;
	outActorParams->mGetActorHelpStringProc				= GetHelpString;
	outActorParams->mCreateActorProc					= CreateActor;
	outActorParams->mDisposeActorProc					= DisposeActor;
	outActorParams->mActivateActorProc					= ActivateActor;
	outActorParams->mHandlePropertyChangeValueProc		= HandlePropertyChangeValue;
	
	// OPTIONAL FUNCTIONS
	outActorParams->mHandlePropertyChangeTypeProc		= HandlePropertyChangeType;
	outActorParams->mHandlePropertyConnectProc			= NULL;
	outActorParams->mPropertyValueToStringProc			= NULL;
	outActorParams->mPropertyStringToValueProc			= NULL;
	outActorParams->mGetActorDefinedAreaProc			= GetActorDefinedArea;
	outActorParams->mDrawActorDefinedAreaProc			= DrawActorDefinedArea;
	outActorParams->mMouseTrackInActorDefinedAreaProc	= NULL;
}

// ---------------------------------------------------------------------------------
//		DrawUnityQuad DEFINES
// ---------------------------------------------------------------------------------
/*
#define		TL			0
#define		TR			1
#define		BL			2
#define		BR			3

#define		VX_X		0
#define		VX_Y		1
#define		TX_X		2
#define		TX_Y		3

void
DrawUnityQuad_Textured(
	const float			inTxCoordRight,
	const float			inTxCoordBottom)
{
	static GLfloat sCoordinates[4][4] = {
		// vertex x/y, texture x/y
		{ -1.0f, -1.0, +0.0f, +0.0f },	// upper left
		{ +1.0f, -1.0, +1.0f, +0.0f },	// upper right
		{ -1.0f, +1.0, +0.0f, +1.0f },	// lower left
		{ +1.0f, +1.0, +1.0f, +1.0f }	// lower right
	};

	sCoordinates[TR][TX_X] = sCoordinates[BR][TX_X] = inTxCoordRight;
	sCoordinates[BL][TX_Y] = sCoordinates[BR][TX_Y] = inTxCoordBottom;
	
	glEnableClientState(GL_VERTEX_ARRAY);
	glVertexPointer(2, GL_FLOAT, 4 * sizeof(GLfloat), &sCoordinates[0][0]);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer(2, GL_FLOAT, 4 * sizeof(GLfloat), &sCoordinates[0][2]);
	
	glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);
}
*/

// ---------------------------------------------------------------------------------
//		ReceiveMessage
// ---------------------------------------------------------------------------------
//	Handles incoming video frame changed messages

static void
ReceiveMessage(
	IsadoraParameters*	ip,
	MessageMask			inMessageMask,
	PortIndex			/* inIndex1 */,
	const MsgData*		inData,
	UInt32				/* inLen */,
	IzRefCon			inRefCon)
{
	ActorInfo*	inActorInfo = (ActorInfo*) inRefCon;
	PluginInfo*	info = (PluginInfo*) inActorInfo->mActorDataPtr;

	if ((inMessageMask & kWantStageAboutToDispose) != 0) {
		// const StageIndexT* stageIndex1 = reinterpret_cast<const StageIndexT*>(inData);
		DisposeVideoConverter(ip, inActorInfo);
		return;
	}

	if (info->mNeedsRender) {

		// call this function when you start processing video; match this
		// with a call to ExitVideoProcessing_ after you're done. This
		// is part of what Isadora does to calculate the LOAD shown in the
		// user interface.
		UInt64 vpStart = EnterVideoProcessing_(ip);
		
		ROIRect roi;
		#if ANALAYZE_PARTIAL_IMAGE
		
			// if the ANALAYZE_PARTIAL_IMAGE switch is true, then calculate
			// the "region of interest", i.e., the area within the source
			// video that we will analyze using the source position (mSrcLeft,
			// mSrcTop) and source sizes (mSrcWidth, mSrcHeight).
			//
			// This can increase efficiency by reducing the amount of data that
			// is transferred from the GPU's video memory to the main memory of
			// the computer
			
			float widthMax = 1.0f - info->mSrcWidth;
			float heightMax = 1.0f - info->mSrcHeight;
			roi.left = widthMax * info->mSrcLeft;
			roi.top = heightMax * info->mSrcTop;
			roi.right = roi.left + info->mSrcWidth;
			roi.bottom = roi.top + info->mSrcHeight;
		
		#else
		
			// if the ANALAYZE_PARTIAL_IMAGE switch is false, then we will
			// always analyze the entire image
			roi.left = 0.0f;
			roi.top = 0.0f;
			roi.right = 1.0f;
			roi.bottom = 1.0f;
		
		#endif
		
		// **************************************************************
		// NOTE: THE NEXT OPERATION IS VERY INEFFICIENT!!!
		// **************************************************************
		//
		// It is very important that you read the READ ME section at the
		// top of this document!
		
		// now convert and crop the source video, whter it be vid-gpu, vid-ci, or vid-cpu, to a bitmap
		Value v;
		info->mBitmapConverter = ConvertVideoStreamToBitmap_(ip, inActorInfo, info->mBitmapConverter, FALSE, &info->mInputBuffer, &roi, &v);
		
		// ensure the bitmap converter exists and that it produced a value result in 'v'
		if (info->mBitmapConverter != NULL && v.type == kData && v.u.data != NULL) {
		
			// get a pointer to the image buffer data
			ImageBufferPtr ib = reinterpret_cast<ImageBufferPtr>(v.u.data);
			
			// then, if the bitmap is valid....
			if (ib->mFormat == kIzzyRGBFormat
			&& ib->mBaseAddress != NULL
			&& ib->mRowBytes > 0
			&& ib->mWidth > 0
			&& ib->mHeight > 0) {
			
				// --------------------------------------------------------------------------------
				// ANALYZE THE BITMAP:
				// --------------------------------------------------------------------------------
				//
				// At this point, you now have the input image/video stored in the main memory
				// of the computer. The image is described by the members of the ImageBuffer struct.
				//
				//     mBitDepth:        the number of bits per pixel; will be 32 for ARGB images
				//     mWidth:           image width in pixels
				//     mHeight:          image height in pixels
				//     mFormat:          will always be kIzzyRGBFormat aka 'BGRA'
				//     mBaseAddress:     the pointer to the first byte of the image
				//
				//     mRowBytes:        the number of bytes per row (also known as 'stride'), which
				//                       is not necessary mWidth * 4. More information below.
				//
				//     mIzzyBitmap:      the underlying bitmap object if one exists: this is a CGImageRef
				//                       on macOS, or an HBITMAP on Windows. If this pointer is NULL, no
				//                       underlying object exists.
				//
				//     mUniqueID:        a 32-bit number that uniquely identifies this bitmap image
				//
				//     mHasAlpha:        will be true if Isadora has sensed that the image has alpha values
				//                       that are not 255 (i.e., opaque).
				//
				//     mMagic:           magic value used to determine the validity of the bitmap
				//
				//     mIsPremultiplied: will be true if the image uses premultiplied alpha.
				//
				//
				// Isadora Bitmap Characteristics:
				//
				// * Isadora bitmaps are always top-down, meaning that the first byte pointed to by
				// mBaseAddress is the top-left corner of the image. The memory layout after conversion
				// will always be BGRA, meaning that the first byte is the blue channel, the second byte
				// is the green channel, and so on. The code below uses the macros ALPHA_, RED_, GREEN_,
				// BLUE_ to pull the individual color components from a 32 bit integer pixels red from the
				// bitmap.
				//
				// * Do not assume that the number of bytes per row is mWidth * 4. You must use the
				// mRowBytes member to ascertain the number of bytes per row! You can see this in the
				// code below in the line:
				//
				// pRowStart = (UInt32*)(((char*) pRowStart + rowBytes));
				//
				// which advances the pRowStart pointer to the next row in the correct manner.
				//
				// --------------------------------------------------------------------------------


				// clear the mNeedsRender, since we're "rendering" the image here
				// (meaning that we're measuring it's color values.)
				info->mNeedsRender = false;
				
				// copy rowBytes to a constant value to potentially increase looping speed
				const UInt32 rowBytes = ib->mRowBytes;
				
				// debugging code checks that the pointers we're using to reference the
				// bitmap data are actually within the bitmap
				#if DEBUG
				UInt32* bufferStart = (UInt32*) ib->mBaseAddress;
				UInt32* bufferEnd = (UInt32*)((char*) ib->mBaseAddress + rowBytes * ib->mHeight);
				#endif
				
				// get starting pixel
				UInt32* pRowStart = (UInt32*)((char*) ib->mBaseAddress);

				const SInt32 xmax = ib->mWidth;
				
				SInt32 ymax = ib->mHeight;
				
				const double totalPixels = xmax * ymax;
				PluginAssert_(ip, totalPixels > 0);

				// a, r, g, and b will be used to accumulate the total value
				// for each color component in the image
				UInt64 a = 0;
				UInt64 r = 0;
				UInt64 g = 0;
				UInt64 b = 0;
				
				// while we have more rows to process
				while (ymax-- > 0) {
					
					// get a pointer to the first pixel
					UInt32* p = pRowStart;
					
					// while we have more pixels in the row to process
					SInt32 xcount = xmax;
					while (xcount-- > 0) {
						
						#if DEBUG
						PluginAssert_(ip, p >= bufferStart && p < bufferEnd);
						#endif
						
						// accumulate the color values for each component
						a += ALPHA_(*p);
						r += RED_(*p);
						g += GREEN_(*p);
						b += BLUE_(*p);
						p++;
					}
					
					// advance the pointer to the next row, using the
					// row bytes value to ensure we skip the right amount
					pRowStart = (UInt32*)(((char*) pRowStart + rowBytes));
				}

				// divide the four accumualted color values by the total
				// number of pixels and divide by 255.0 to get a value
				// between 0.0 and 1.0.
				double fa = ((double) a / totalPixels) / 255.0;
				double fr = ((double) r / totalPixels) / 255.0;
				double fg = ((double) g / totalPixels) / 255.0;
				double fb = ((double) b / totalPixels) / 255.0;

				// if the color values are premultiplied, we need to
				// unpremutiply them to get the true intensity.
				if (ib->mIsPremultiplied) {
					if (fa > 0.0) {
						fr = fr / fa;
						fg = fg / fa;
						fb = fb / fa;
					} else {
						fr = fg = fb = 0.0;
					}
				}
				
				// scale the final output values so they are between 0 and 255
				const UInt8 aOut = UInt8(round(fa * 255.0));
				const UInt8 rOut = UInt8(round(fr * 255.0));
				const UInt8 gOut = UInt8(round(fg * 255.0));
				const UInt8 bOut = UInt8(round(fb * 255.0));
				
				v.type = kInteger;

				// set the actor's outputs
				v.u.ivalue = rOut;
				SetOutputPropertyValue_(ip, inActorInfo, kOutputRed, &v);
				v.u.ivalue = gOut;
				SetOutputPropertyValue_(ip, inActorInfo, kOutputGreen, &v);
				v.u.ivalue = bOut;
				SetOutputPropertyValue_(ip, inActorInfo, kOutputBlue, &v);
				v.u.ivalue = aOut;
				SetOutputPropertyValue_(ip, inActorInfo, kOutputAlpha, &v);
				
				/*
				ColorRef colorRef = RGBToColorRef(rOut, gOut, bOut);
				v.type = kFloat;
				v.u.fvalue = ColorRefToFloat(colorRef);
				SetOutputPropertyValue_(ip, inActorInfo, kOutputColor, &v);
				*/

				// and then use the color components to create a color ref
				// that we can output a average color value that can be
				// passed on to actors that accept the color data type as
				// an input
				ColorRef colorRef = ARGBToColorRef(aOut, rOut, gOut, bOut);
				v.type = kInteger;
				v.u.ivalue = colorRef;
				SetOutputPropertyValue_(ip, inActorInfo, kOutputColor, &v);

			}
		}
		
		ExitVideoProcessing_(ip, vpStart);
	}
}
