// ===========================================================================
//	IsadoraPlugin				   2001 Mark F. Coniglio. All rights reserved.
// ===========================================================================

#include <string.h>
#include <stdio.h>
#include <float.h>
#include <ctype.h>
#include <vector>
#include <string>
#include <fstream>

#include "IsadoraPluginPrefix.h"

#include "IsadoraCallbacks.h"

#if TARGET_OS_MAC
const char kSeparatorChar = '/';
#elif TARGET_OS_WIN
const char kSeparatorChar = '\\';
#endif

// ---------------------------------------------------------------------------------
// MacOS Specific
// ---------------------------------------------------------------------------------
#if TARGET_OS_MAC
#define EXPORT_
#endif

#if FALSE && DEBUG && TARGET_OS_MAC
	#define	log_info	printf
#else
	#ifdef log_info
	#undef log_info
	#endif
	#define	log_info(...) { }
#endif

// ---------------------------------------------------------------------------------
// Win32  Specific
// ---------------------------------------------------------------------------------
#if TARGET_OS_WIN
	#define EXPORT_ __declspec(dllexport)
#endif

// ---------------------------------------------------------------------------------
//	Exported Function Definitions
// ---------------------------------------------------------------------------------

#ifdef __cplusplus
extern "C" {
#endif

EXPORT_ void GetActorInfo(void* inParam, ActorInfo* outActorParams);

#ifdef __cplusplus
}
#endif


// ---------------------------------------------------------------------------------
//	Function Prototypes
// ---------------------------------------------------------------------------------
static void
ReceiveMessage(
	IsadoraParameters*	ip,
	MessageMask			inMessageMask,
	PortIndex			inPortIndex1,
	const MsgData*		inData,
	UInt32				inLen,
	long				inRefCon);
	
// ---------------------------------------------------------------------------------
// PluginInfo struct
// ---------------------------------------------------------------------------------
// This structure neeeds to contain all variables used by your plugin. Memory for
// this struct is allocated during the CreateActor function, and disposed during
// the DisposeActor function.

typedef struct {
	ActorInfo*	mActorInfoPtr;
	char		mFilePath[512];
} PluginInfo;

// A handy macro for casting the mActorDataPtr to PluginInfo*
#if __cplusplus
#define	GetPluginInfo_(actorDataPtr)		static_cast<PluginInfo*>((actorDataPtr)->mActorDataPtr);
#else
#define	GetPluginInfo_(actorDataPtr)		(PluginInfo*)((actorDataPtr)->mActorDataPtr);
#endif

// ---------------------------------------------------------------------------------
//	Constants
// ---------------------------------------------------------------------------------
//	Defines various constants used throughout the plugin.

// * GROUP ID
static const OSType	kActorClass 	= kGroupText;

// * PLUGIN IN
static const OSType	kActorID		= FOUR_CHAR_CODE('FILE');

// * ACTOR NAME
static const char* kActorName		= "File Reader";

// * PROPERTY DEFINITION STRING
static const char* sPropertyDefinitionString =

		"INPROP 	file_path	fpat	string		text				*		*		your-file-name.txt\r"
		"INPROP		read		read	bool		trig				0		1		0\r"

		"OUTPROP	text		text	string		text				*		*		\r"
		"OUTPROP	length		slen	int			number				0		*		0\r";

// * Property Index Constants
enum
{
	kInputFilePath = 1,
	kInputRead,
	
	kOutputText = 1,
	kOutputTextLength
};

// * Help Strings
const char* sHelpStrings[] =
{
	// INPUTS
	"Reads a text file and outputs its contents.",
	
	"The file name, or the full or partial path name, of the text file to read when the 'read' input is triggered.\r"
	
		"The best practice is place the file you wish to read in the same folder as your Isadora (.izz) file. If you "
		"do this, then you need only to enter the file name and extension for this property, e.g., file-name.txt.\r"
	
		"Otherwise you can specify a full or partial path name. Full path names start with a '/' on macOS, or 'X:\' "
		"on Windows, where X is a drive letter. On macOS, if you wish to read from a drive other than the main hard drive, "
		"the full path name will start with '/Volumes/NAME' where NAME is the name of the drive.\r"
	
		"Partial path names are specified relative to the Isadora document, and can use the special code '..' to "
		"designate the folder that is one above the current directory. For example, if your file is in folder called "
		"'myfolder' that is in the same folder as your Isadora document, you could use the partial path "
		"'myfolder/file-name.txt. If the file is in the directory above your Isadora document, then you could "
		"enter '../file-name.txt.",
	
	"Reads the contents of the file specified by the 'file path' property and send the output to the 'text' "
		"output.",
	
	// OUTPUTS
	
	"The text contained in the file specified by the 'file path' property.",
	
	"The number of bytes in the text file."
};

// ---------------------------------------------------------------------------------
//		 IsRelativePathName
// ---------------------------------------------------------------------------------
static bool
IsRelativePathName(
	const char*		c)
{
	bool isRelativePath = true;
	
	#if TARGET_OS_MAC
	if (strlen(c) >= 2) {
		if (*c == kSeparatorChar) {
			isRelativePath = false;
		}
	}
	#elif TARGET_OS_WIN
	if (strlen(c) >= 3) {
		if ( ((*c >= 'a' && *c <= 'z') || (*c >= 'A' && *c <= 'Z')) && *(c+1) == ':' && *(c+2) == kSeparatorChar) {
			isRelativePath = false;
		}
	}
	#endif
	
	return isRelativePath;
}

// ---------------------------------------------------------------------------------
//		 ReadFile
// ---------------------------------------------------------------------------------

static void
ReadFile(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo)
{
	PluginInfo* info = (PluginInfo*) inActorInfo->mActorDataPtr;
	
	char fpBuf[4096];
	Boolean success = GetUTF8PathToCurrentDocument_(ip, inActorInfo->mThing, sizeof(fpBuf), fpBuf);
	if (success && IsRelativePathName(info->mFilePath)) {
		// relative path: the full path is the document path + the supplied relative path
		std::strncat_s(fpBuf, info->mFilePath, sizeof(fpBuf));
	} else {
		// relative path: the full path is the supplied path
		std::strncpy_s(fpBuf, info->mFilePath, sizeof(fpBuf));
	}
	
	#if TARGET_OS_MAC
	
		// on macOS, we can use the UTF8 path
		std::ifstream f(fpBuf, std::ios::binary);
	
	#elif TARGET_OS_WIN
	
		// on Windows, we must convert the path to a wide character string
	
		// determine number of WCHARs required and allocate buffer
		int count = MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, fpBuf, (int) strlen(fpBuf), NULL, 0);
		size_t bufferSize = count + 1;
		WCHAR* wstr = (WCHAR*) malloc(bufferSize * sizeof(WCHAR));
		
		// attempt to convert UTF-8 path to WCHAR string
		int conversionResult = ::MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, fpBuf, (int) strlen(fpBuf), wstr, (int) bufferSize);
		if (conversionResult >= 0 && conversionResult < bufferSize) {
			wstr[conversionResult] = 0;
		} else {
			wstr[0] = 0;
		}
	
		std::ifstream f(wstr, std::ios::binary);
	
		free(wstr);

	#endif
	
	std::vector<char> buffer((std::istreambuf_iterator<char>(f)), std::istreambuf_iterator<char>());
	
	// handle any line ending type:
	// mac = CR only
	// win = CR+LF
	// unix = LF only
	
	size_t i = 0;
	while (i < buffer.size()) {
		
		// if we find a LF (without a preceding CR) this is
		// a unix text file -- replace the LF with a CR
		
		if (buffer[i] == 0x0A) {
			
			buffer[i] = 0x0D;
			
		// if the character is a CR it could be a mac text
		// file (CR only) or a windows text file (CR+LF)
		
		} else if (buffer[i] == 0x0D) {
		
			// if we're not a the end of the file and the next
			// character is a LF, then we have a CR-LF pair;
			// delete the LF
			if ((i+1) < buffer.size() && buffer[i+1] == 0x0A) {
				// erase the linefeed from the vector
				buffer.erase(buffer.begin() + (i+1));
			
			// otherwise it's CR only -- just leave it be
			} else {
				i++;
			}
		
		// otherewise it's some other character -- just leave it be
		} else {
			i++;
		}
	}
	
	// null terminate buffer
	buffer.push_back(0);
	
	// std::string str( (std::istreambuf_iterator<char>(f)), std::istreambuf_iterator<char>() );
	
	Value v = { 0, { 0 } };
	AllocateValueString_(ip, &buffer[0], &v);
	SetOutputPropertyValue_(ip, inActorInfo, kOutputText, &v);
	ReleaseValueString_(ip, &v);

	v.type = kInteger;
	v.u.ivalue = buffer.size();
	SetOutputPropertyValue_(ip, inActorInfo, kOutputTextLength, &v);
}

// ---------------------------------------------------------------------------------
//		 CreateActor
// ---------------------------------------------------------------------------------
static void
CreateActor(
	IsadoraParameters*	ip,
	ActorInfo*			ioActorInfo)
{
	PluginInfo* info = (PluginInfo*) IzzyMallocClear_(ip, sizeof(PluginInfo));
	PluginAssert_(ip, info != nil);
	
	ioActorInfo->mActorDataPtr = info;
	info->mActorInfoPtr = ioActorInfo;
	info->mFilePath[0] = 0;
}

// ---------------------------------------------------------------------------------
//		 DisposeActor
// ---------------------------------------------------------------------------------
static void
DisposeActor(
	IsadoraParameters*	ip,
	ActorInfo*			ioActorInfo)
{
//	PluginInfo* info = (PluginInfo*) ioActorInfo->mActorDataPtr;
	
	
	PluginAssert_(ip, ioActorInfo->mActorDataPtr != nil);
	IzzyFree_(ip, ioActorInfo->mActorDataPtr);
}

// ---------------------------------------------------------------------------------
//		 ActivateActor
// ---------------------------------------------------------------------------------

static void
ActivateActor(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	Boolean				inActivate)
{
	PluginInfo* info = (PluginInfo*) inActorInfo->mActorDataPtr;

	if (inActivate) {
	
		info->mFilePath[0] = 0;
		
		if (ip->mCallbacks->mCallbackVersionNumber >= kIsadoraCallbacksVersion_308b39) {
			SetPropertyFilePathModeProc_(ip, inActorInfo, kInputProperty, kInputFilePath, kPropertyFilePath_PathRead);
		}

	} else {
		
	}
}

// ---------------------------------------------------------------------------------
//		 GetParameterString
// ---------------------------------------------------------------------------------

static const char*
GetParameterString(
	IsadoraParameters*	/* ip */,
	ActorInfo*			/* inActorInfo */)
{
	return sPropertyDefinitionString;
}

// ---------------------------------------------------------------------------------
//		 GetHelpString
// ---------------------------------------------------------------------------------

static void
GetHelpString(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	PropertyType		inPropertyType,
	PropertyIndex		inPropertyIndex1,
	char*				outParamaterString,
	UInt32				inMaxCharacters)
{
	const char* helpstr = nil;
	
	UInt32 index1 = PropertyTypeAndIndexToHelpIndex_(ip, inActorInfo, inPropertyType, inPropertyIndex1);

	helpstr = sHelpStrings[index1];

	strncpy(outParamaterString, helpstr, inMaxCharacters);
}
	
// ---------------------------------------------------------------------------------
//		 HandlePropertyChangeValue	[INTERRUPT SAFE]
// ---------------------------------------------------------------------------------

static void
HandlePropertyChangeValue(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	PropertyIndex		inPropertyIndex1,
	ValuePtr			/* inOldValue */,
	ValuePtr			inNewValue,
	Boolean				inInitializing)
{

	PluginInfo* info = (PluginInfo*) inActorInfo->mActorDataPtr;	

	Value outputValue;
	outputValue.type = kFloat;
	
	
	bool readFile = false;
	
	switch (inPropertyIndex1) {

	case kInputFilePath:
		if (inNewValue->type == kString) {
			PluginAssert_(ip, strlen(inNewValue->u.str->strData)+1 <= sizeof(info->mFilePath));
			strncpy(info->mFilePath, inNewValue->u.str->strData, sizeof(info->mFilePath)-1);
			info->mFilePath[ sizeof(info->mFilePath)-1] = 0;
		}
		readFile = true;
		break;

	case kInputRead:
		readFile = true;
		break;
	}

	if (readFile) {
		ReadFile(ip, inActorInfo);
	}
}


// ---------------------------------------------------------------------------------
//		 GetActorDefinedArea
// ---------------------------------------------------------------------------------

static ActorPictInfo	gPictInfo = { false, nil, nil, 0, 0 };

static Boolean
GetActorDefinedArea(
	IsadoraParameters*			ip,			
	ActorInfo*					inActorInfo,
	SInt16*						outTopAreaWidth,
	SInt16*						outTopAreaMinHeight,
	SInt16*						outBotAreaHeight,
	SInt16*						outBotAreaMinWidth)
{
	if (gPictInfo.mInitialized == false) {
		PrepareActorDefinedAreaPict_(ip, inActorInfo, 0, &gPictInfo);
	}
	
	*outTopAreaWidth = gPictInfo.mWidth;
	*outTopAreaMinHeight = gPictInfo.mHeight;
	*outBotAreaHeight = 0;
	*outBotAreaMinWidth = 0;

	
	return true;
}

// ---------------------------------------------------------------------------------
//		 DrawActorDefinedArea
// ---------------------------------------------------------------------------------

static void
DrawActorDefinedArea(
	IsadoraParameters*			ip,
	ActorInfo*					inActorInfo,
	void*						/* inDrawingContext */,
	ActorDefinedAreaPart		inActorDefinedAreaPart,
	ActorAreaDrawFlagsT			/* inAreaDrawFlags */,
	Rect*						inADAArea,
	Rect*						/* inUpdateArea */,
	Boolean						inSelected)
{
	if (inActorDefinedAreaPart == kActorDefinedAreaTop && gPictInfo.mInitialized) {
		DrawActorDefinedAreaPict_(ip, inActorInfo, inSelected, inADAArea, &gPictInfo);
	}
}
	
// ---------------------------------------------------------------------------------
//		 GetActorInfo
// ---------------------------------------------------------------------------------
//	This is function is called by to get the actor's class and ID, and to get
//	pointers to the all of the plugin functions declared locally.
//
//	All members of the ActorInfo struct pointed to by outActorParams have been
//	set to 0 on entry. You only need set functions defined by your plugin
//	
EXPORT_ void
GetActorInfo(
	void*				/* inParam */,
	ActorInfo*			outActorParams)
{
	// REQUIRED information
	outActorParams->mActorName							= kActorName;
	outActorParams->mClass								= kActorClass;
	outActorParams->mID									= kActorID;
	outActorParams->mCompatibleWithVersion				= kActorCompatilibilityVersion_v1;
	outActorParams->mActorFlags							= kActorFlags_Plugin_CheckForUpdates;
	
	// REQUIRED functions
	outActorParams->mGetActorParameterStringProc		= GetParameterString;
	outActorParams->mGetActorHelpStringProc				= GetHelpString;
	outActorParams->mCreateActorProc					= CreateActor;
	outActorParams->mDisposeActorProc					= DisposeActor;
	outActorParams->mActivateActorProc					= ActivateActor;
	outActorParams->mHandlePropertyChangeValueProc		= HandlePropertyChangeValue;
	
	// OPTIONAL FUNCTIONS
	outActorParams->mHandlePropertyConnectProc			= NULL;
	outActorParams->mPropertyValueToStringProc			= NULL;
	outActorParams->mPropertyStringToValueProc			= NULL;
	outActorParams->mGetActorDefinedAreaProc			= GetActorDefinedArea;
	outActorParams->mDrawActorDefinedAreaProc			= DrawActorDefinedArea;
	outActorParams->mMouseTrackInActorDefinedAreaProc	= NULL;
}
