// ===========================================================================
//	IsadoraPlugin				   2001 Mark F. Coniglio. All rights reserved.
// ===========================================================================

// STANDARD INCLUDES
#include <string.h>
#include <stdio.h>
#include <stdint.h>
#include <float.h>
#include <ctype.h>
#include <vector>
#include <string>
#include <fstream>

// INCLUDE JSON
#include "nlohmann/json.hpp"
using json = nlohmann::json;

// ISADORA INCLUDES
#include "IsadoraPluginPrefix.h"
#include "IsadoraCallbacks.h"

#define MUTABLE_ITEMS		FALSE

// ---------------------------------------------------------------------------------
// MacOS Specific
// ---------------------------------------------------------------------------------
#if TARGET_OS_MAC
#define EXPORT_
#endif

#if FALSE && DEBUG && TARGET_OS_MAC
	#define	log_info	printf
#else
	#ifdef log_info
	#undef log_info
	#endif
	#define	log_info(...) { }
#endif

// ---------------------------------------------------------------------------------
// Win32  Specific
// ---------------------------------------------------------------------------------
#if TARGET_OS_WIN
	#define EXPORT_ __declspec(dllexport)
#endif

// ---------------------------------------------------------------------------------
//	Exported Function Definitions
// ---------------------------------------------------------------------------------

#ifdef __cplusplus
extern "C" {
#endif

EXPORT_ void GetActorInfo(void* inParam, ActorInfo* outActorParams);

#ifdef __cplusplus
}
#endif


static Boolean gInsideHandlePropertyChangeType = false;

// ---------------------------------------------------------------------------------
//	Function Prototypes
// ---------------------------------------------------------------------------------
static void
ReceiveMessage(
	IsadoraParameters*	ip,
	MessageMask			inMessageMask,
	PortIndex			inPortIndex1,
	const MsgData*		inData,
	UInt32				inLen,
	long				inRefCon);
	
// ---------------------------------------------------------------------------------
// PluginInfo struct
// ---------------------------------------------------------------------------------
// This structure neeeds to contain all variables used by your plugin. Memory for
// this struct is allocated during the CreateActor function, and disposed during
// the DisposeActor function.

#define kMaxOutputNeedsUpdateFlags	100

typedef struct {
	ActorInfo*	mActorInfoPtr;
	json*		mJSON;
	bool		mOutputNeedsUpdate[kMaxOutputNeedsUpdateFlags];
} PluginInfo;

// A handy macro for casting the mActorDataPtr to PluginInfo*
#if __cplusplus
#define	GetPluginInfo_(actorDataPtr)		static_cast<PluginInfo*>((actorDataPtr)->mActorDataPtr);
#else
#define	GetPluginInfo_(actorDataPtr)		(PluginInfo*)((actorDataPtr)->mActorDataPtr);
#endif

// ---------------------------------------------------------------------------------
//	Constants
// ---------------------------------------------------------------------------------
//	Defines various constants used throughout the plugin.

// * GROUP ID
static const OSType	kActorClass 	= kGroupText;

// * PLUGIN IN
static const OSType	kActorID		= FOUR_CHAR_CODE('JSON');

// * ACTOR NAME
static const char* kActorName		= "JSON Parser";

// * PROPERTY DEFINITION STRING
static const char* sPropertyDefinitionString =

		"INPROP 	json_in		fpat	string		text				*		*		\r"
		"INPROP 	selectors	selc	int			number				1		100		1\r"
		"INPROP 	select_1	i001	string		text				*		*		\r"

		"OUTPROP	error		erro	int			propdef				0		1		0\r"
		"OUTPROP	elements	ecnt	int			number				*		*		0\r"
		"OUTPROP	element_1	o001	string		text				*		*		\r";

// * Property Index Constants
enum
{
	kInputJSON = 1,
	kInputItemCount,
	kInput1Base,
	
	kFixedInputPropertyCountBefore	= 2,
	kFixedInputPropertyCountAfter	= 0,
	kTotalFixedInputPropertyCount	= (kFixedInputPropertyCountBefore + kFixedInputPropertyCountAfter),
	kTotalInputPropertyCount		= (kTotalFixedInputPropertyCount+1),
	kFirstVariableInputProperty		= (kFixedInputPropertyCountBefore+1),

	kOutputError = 1,
	kOutputListSize,
	
	kFixedOutputPropertyCount		= 2,
	kFirstVariableOutputProperty	= (kFixedOutputPropertyCount+1)
};

// * Help Strings
const char* sHelpStrings[] =
{
	// INPUTS
	"Parses JSON data contained in a text string.",
	
	"A text string containing value JSON data.",
	
	"The number of input selector strings",
	
	"A string in the format xxx:yyy:zzz that selects a data member for the corresponding output, where xxx, yyy "
		"and zzz can either be a key value or a numeric index. Items with only numeric digits are interpreted as an "
		"index. Items with non-numeric characters are interpreted as a key. Keys that conists only of numeric "
		"characters or that contain a colon must be enclosed in double quotes.\r\r"
		"For example, the selector mark:3:leg 1 would attempt to get the top level element named 'mark', "
		"the third item inside of 'mark', and then a keyed value called 'leg 1' within that element.",
	
	// OUTPUTS
	
	"Outputs 'no' if the JSON data is valid and was sucessfully parsed. Otherwise outputs 'yes'.",
	
	"Outputs the number of top level elements in the object. Sent whenever the 'json in' input changes and the JSON is parsed.",
	
	"An element of the JSON input, a specified by the corresponding selector property."
};

// ---------------------------------------------------------------------------------
//		 CreatePropertyID	[INTERRUPT SAFE]
// ---------------------------------------------------------------------------------

inline OSType
CreatePropertyID(
	IsadoraParameters*	ip,
	const char*			inRateBase,
	SInt32				inIndex)
{
	PluginAssert_(ip, inRateBase[0] != 0);
	PluginAssert_(ip, inIndex >= 0 && inIndex < 999);
	
	OSType	result = (((UInt32) inRateBase[0]) << 24);
	
	// in index is between 000 and 999
	if (inIndex >= 0 && inIndex < 999) {
	
		UInt32 indexHS = inIndex / 100;
		UInt32 indexMS = (inIndex % 100) / 10;
		UInt32 indexLS = inIndex % 10;
	
		result |= ( (((UInt32) (indexHS + '0')) << 16) | (((UInt32) (indexMS + '0')) << 8) | (((UInt32) (indexLS + '0')) << 0) );
	
	} else {
		PluginAssert_(ip, false);
	}
	
	return result;
}

// ---------------------------------------------------------------------------------
//		 IsRelativePathName
// ---------------------------------------------------------------------------------
static bool
IsRelativePathName(
	const char*		c)
{
	bool isRelativePath = true;
	
	#if TARGET_OS_MAC
	if (strlen(c) >= 2) {
		if (*c == '/') {
			isRelativePath = false;
		}
	}
	#elif TARGET_OS_WIN
	if (strlen(c) >= 3) {
		if ( ((*c >= 'a' && *c <= 'z') || (*c >= 'A' && *c <= 'Z')) && *(c+1) == ':' && *(c+2) == '\\') {
			isRelativePath = false;
		}
	}
	#endif
	
	return isRelativePath;
}

// ---------------------------------------------------------------------------------
//		 ReadFile
// ---------------------------------------------------------------------------------

static void
LoadJSON(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	const char*			inJSONString)
{
	PluginInfo* info = (PluginInfo*) inActorInfo->mActorDataPtr;
	
	Value error;
	error.type = kInteger;
	error.u.ivalue = 0;
	
	Value listSize;
	listSize.type = kInteger;
	listSize.u.ivalue = 0;
	
	try {
	
		info->mJSON = new json();
		
		memset(info->mOutputNeedsUpdate, 1, sizeof(info->mOutputNeedsUpdate));
		
		*(info->mJSON) = json::parse(std::string(inJSONString));
		
		listSize.u.ivalue = info->mJSON->size();

	} catch (const nlohmann::detail::parse_error& perr) {
		
		error.u.ivalue = 1;
		PluginLogInfo_(ip, "JSON Parse Error:  %s\n", perr.what());
		
	} catch (...) {
	
		error.u.ivalue = 1;
		PluginLogInfo_(ip, "JSON Errror\n");
	}

	SetOutputPropertyValue_(ip, inActorInfo, kOutputError, &error);
	SetOutputPropertyValue_(ip, inActorInfo, kOutputListSize, &listSize);
}

// ---------------------------------------------------------------------------------
//		 CreateActor
// ---------------------------------------------------------------------------------
static void
CreateActor(
	IsadoraParameters*	ip,
	ActorInfo*			ioActorInfo)
{
	PluginInfo* info = (PluginInfo*) IzzyMallocClear_(ip, sizeof(PluginInfo));
	PluginAssert_(ip, info != nil);
	
	ioActorInfo->mActorDataPtr = info;
	info->mActorInfoPtr = ioActorInfo;
}

// ---------------------------------------------------------------------------------
//		 DisposeActor
// ---------------------------------------------------------------------------------
static void
DisposeActor(
	IsadoraParameters*	ip,
	ActorInfo*			ioActorInfo)
{
//	PluginInfo* info = (PluginInfo*) ioActorInfo->mActorDataPtr;
	
	
	PluginAssert_(ip, ioActorInfo->mActorDataPtr != nil);
	IzzyFree_(ip, ioActorInfo->mActorDataPtr);
}

// ---------------------------------------------------------------------------------
//		 ActivateActor
// ---------------------------------------------------------------------------------

static void
ActivateActor(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	Boolean				inActivate)
{
	PluginInfo* info = (PluginInfo*) inActorInfo->mActorDataPtr;

	if (inActivate) {

		info->mJSON = NULL;
		
	} else {
		
		if (info->mJSON != NULL) {
			delete info->mJSON;
		}
	}
}

// ---------------------------------------------------------------------------------
//		 GetParameterString
// ---------------------------------------------------------------------------------

static const char*
GetParameterString(
	IsadoraParameters*	/* ip */,
	ActorInfo*			/* inActorInfo */)
{
	return sPropertyDefinitionString;
}

// ---------------------------------------------------------------------------------
//		 GetHelpString
// ---------------------------------------------------------------------------------

static void
GetHelpString(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	PropertyType		inPropertyType,
	PropertyIndex		inPropertyIndex1,
	char*				outParamaterString,
	UInt32				inMaxCharacters)
{
	const char* helpstr = nil;
	
	// if the input the user is asking about is
	// past the end of the fixed properties, then
	// we force it to the first fixed property
	if (inPropertyType == kInputProperty) {
	
		UInt32 inPropCount;
		GetPropertyCount_(ip, inActorInfo, kInputProperty, &inPropCount);
		UInt32 totalVariablePropCount = inPropCount - (kFixedInputPropertyCountBefore + kFixedInputPropertyCountAfter);
		
		if (inPropertyIndex1 < kFirstVariableInputProperty) {
			// do nothing -- index is arleady correct
		} else if (inPropertyIndex1 <= kFixedInputPropertyCountBefore + totalVariablePropCount) {
			inPropertyIndex1 = kFirstVariableInputProperty;
		} else {
			inPropertyIndex1 = inPropertyIndex1 - totalVariablePropCount + 1;
		}
		
	} else if (inPropertyType == kOutputProperty) {
		if (inPropertyIndex1 > kFixedOutputPropertyCount) {
			inPropertyIndex1 = kFirstVariableOutputProperty;
		}
		inPropertyIndex1 += kTotalInputPropertyCount;
	}
	
	helpstr = sHelpStrings[inPropertyIndex1];
	
	strncpy(outParamaterString, helpstr, inMaxCharacters);
}
	
// ---------------------------------------------------------------------------------
//		 HandlePropertyChangeValue	[INTERRUPT SAFE]
// ---------------------------------------------------------------------------------

class CElement
{
public:

	const size_t kInvalidElement = (size_t) -1;
	
	enum ElementType {
		kIndex,
		kKey
	};
	
	CElement(
		size_t		inIndex) :
			mType(kIndex),
			mIndex(inIndex),
			mKey()
	{
	}
	
	CElement(
		std::string		inKey) :
			mType(kKey),
			mIndex(kInvalidElement),
			mKey(inKey)
	{
	}

	ElementType mType;
	size_t		mIndex;
	std::string	mKey;
	
};


static const char*
ParseElement(
	const char*		c,
	bool&			outError,
	bool&			outNumeric,
	std::string&	outToken)
{
	outError = false;
	outToken.clear();
	
	char buffer[1024] = { 0 };
	size_t bufferAvail = 1024;
	char* d = buffer;
	
	// determine if token begins with
	// a double quote
	bool insideDoubleQuote = false;
	if (*c == '\"') {
		insideDoubleQuote = true;
		c++;
	}
	
	bool isNumeric = true;
	
	while (*c != 0 && *c != ':' && bufferAvail > 0) {
	
		bool charIsOK = false;
		
		if (*c < '0' || *c > '9') {
			isNumeric = false;
		}
		
		// if we're inside a quoted string,
		// and we encounter a double quote
		// then it is an error if it is not
		// the end of a string, as marked by
		// a null character or a colon
		if (insideDoubleQuote && *c == '"') {
			
			// skip the double quote
			c++;
			
			if (*c == 0 || *c == ':') {
				charIsOK = true;
			} else {
				outError = true;
			}
			
		// allow escapes of colons using backslash + colon
		} else if (*c == '\\' && *c == ':') {
			*(d++) = ':';
			bufferAvail -= 1;
			charIsOK = true;
			c++;
		// allow escapes of double quotes using backslash + double quote
		} else if (*c == '\\' && *c == '"') {
			*(d++) = '"';
			bufferAvail -= 1;
			charIsOK = true;
			c++;
		
		// otherwise, the character is something else and we keep going
		} else {
			*(d++) = *c;
			bufferAvail -= 1;
			charIsOK = true;
			c++;
		}
		
		if (!charIsOK)
			break;
	}
	
	*d = 0;
	
	if (!outError) {
		outNumeric = isNumeric;
		std::string str(buffer);
		outToken = str;
	}
	
	return c;
}

static std::vector<CElement>
ParseElementString(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	const Value*		inValue)
{
	std::vector<CElement> list;
	
	if (inValue->type == kString && inValue->u.str != NULL) {
	
		const char* c = inValue->u.str->strData;
		
		// skip leading whitespace
		while (*c != 0 && isspace(*c)) {
			c++;
		}
		
		bool error = false;
		while (*c != 0 && !error) {
		
			bool isNumeric = false;
			std::string token;
			c = ParseElement(c, error, isNumeric, token);
			
			if (!error) {
				if (isNumeric) {
					list.push_back(CElement(atoi(token.c_str())));
				} else {
					list.push_back(CElement(token));
				}
				
				if (*c == ':') {
					c++;
				}
				
			} else {
				list.clear();
			}
		}
		
		#if DEBUG
		if (!error) {
			PluginLogInfo_(ip, "--------------------------------------------\n");
			PluginLogInfo_(ip, "| Parsed Selector: %s\n", inValue->u.str->strData);
			PluginLogInfo_(ip, "--------------------------------------------\n");
			std::vector<CElement>::iterator i;
			for (i=list.begin(); i<list.end(); i++) {
				CElement& e = *i;
				if (e.mType == CElement::kIndex) {
					PluginLogInfo_(ip, "-- Index = %d\n", int(e.mIndex));
				} else if (e.mType == CElement::kKey) {
					PluginLogInfo_(ip, "-- Key   = '%s'\n", e.mKey.c_str());
				}
			}
		}
		#endif
	}
	
	return list;
}

static json
GetJSONElement(
	IsadoraParameters*				ip,
	ActorInfo*						inActorInfo,
	const std::vector<CElement>&	inList,
	bool&							outError)
{
	PluginInfo* info = (PluginInfo*) inActorInfo->mActorDataPtr;

	std::vector<CElement>::const_iterator i;
	
	json element(*info->mJSON);
	
	outError = false;
	for (i=inList.begin(); !outError && i<inList.end(); i++) {
		
		try {
			if (i->mType == CElement::kIndex) {
				element = element.at(i->mIndex);
			} else if (i->mType == CElement::kKey) {
				element = element[i->mKey];
			}
		
		} catch (const nlohmann::detail::parse_error& perr) {
			
			outError = true;
			PluginLogInfo_(ip, "JSON Get Element Error: %s\n", perr.what());
			
		} catch (...) {
		
			outError = true;
			PluginLogInfo_(ip, "JSON Get Element Error: ???\n");
		}
		
	}
	
	return element;
}

// ---------------------------------------------------------------------------------
//		 HandlePropertyChangeValue	[INTERRUPT SAFE]
// ---------------------------------------------------------------------------------

static void
ProcessInputValue(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	PropertyIndex		inOutputPropertyIndex1,
	const ValuePtr		inValue)
{
	bool error = false;
	std::vector<CElement> elements = ParseElementString(ip, inActorInfo, inValue);
	if (elements.size() > 0) {
		json element = GetJSONElement(ip, inActorInfo, elements, error);
		Value v = { 0, { 0 } };
		std::string str;
		if (!error) {
			char buffer[256];
			if (element.is_string()) {
				str = element;
			} else if (element.is_number_integer()) {
				int intValue = element;
				snprintf(buffer, sizeof(buffer), "%d", intValue);
				str = buffer;
			} else if (element.is_number_float()) {
				double floatValue = element;
				snprintf(buffer, sizeof(buffer), "%f", floatValue);
				str = buffer;
			} else if (element.is_boolean()) {
				bool boolValue = element;
				if (boolValue) {
					str = "true";
				} else {
					str = "false";
				}
			} else if (element.is_object()) {
				str = element.dump();
			}
			AllocateValueString_(ip, str.c_str(), &v);
			SetOutputPropertyValue_(ip, inActorInfo, inOutputPropertyIndex1, &v);
			ReleaseValueString_(ip, &v);
		} else {
			AllocateValueString_(ip, str.c_str(), &v);
			SetOutputPropertyValue_(ip, inActorInfo, inOutputPropertyIndex1, &v);
			ReleaseValueString_(ip, &v);
		}
	}
}

// ---------------------------------------------------------------------------------
//		 HandlePropertyChangeValue	[INTERRUPT SAFE]
// ---------------------------------------------------------------------------------

static void
HandlePropertyChangeValue(
	IsadoraParameters*	ip,
	ActorInfo*			inActorInfo,
	PropertyIndex		inPropertyIndex1,
	ValuePtr			/* inOldValue */,
	ValuePtr			inNewValue,
	Boolean				inInitializing)
{
	PluginInfo* info = (PluginInfo*) inActorInfo->mActorDataPtr;
	
	UInt32 propCount;
	IzzyError err = GetPropertyCount_(ip, inActorInfo, kInputProperty, &propCount);
	PluginAssert_(ip, err == kIzzyNoError && propCount >= kTotalInputPropertyCount);
	
	// number of variable inputs currently in the actor
	UInt32 variableInputCount = propCount - kTotalFixedInputPropertyCount;

	UInt16 item = 0;
	// if the input property index is between the base of the input values
	// and that base plus the number of value inputs, then we are referring
	// to one of the input values
	if (inPropertyIndex1 >= kInput1Base && inPropertyIndex1 < kInput1Base + variableInputCount) {
		item = inPropertyIndex1 - kInput1Base;
		inPropertyIndex1 = kInput1Base;
	// if the input property index is greater than or equal to the input value
	// base and the number of variable properties, then we are referring to
	// one of the remaining input and must adjust the input property
	// index accordingly
	} else if (inPropertyIndex1 >= kInput1Base + variableInputCount) {
		inPropertyIndex1 -= (variableInputCount - 1);
	}

	switch (inPropertyIndex1) {

	case kInputJSON:
		if (inNewValue->type == kString) {
		
			// this is the structr
			struct MyStruct{
				Value mMyString;
			}
			
			// when you receive a value
			if (myStruct.mMyString.str != NULL) {
				ReleaseValueString_(ip, &myStruct.mMyString);
			}
			AllocateValueString_(ip, inNewValue->u.str->strData, &myStruct.mMyString);
			
			// in the deaactivate routing
			if (myStruct.mMyString.str != NULL) {
				ReleaseValueString_(ip, &myStruct.mMyString);
			}

			// this is the structr
			struct MyStruct{
				std::string* myString;
			}
			
			// when you receive a value
			if (myStruct->mString == NULL) {
				myStruct->mString = new std::string(inNewValue->u.str->strData);
			} else {
				*(myStruct->mString) = inNewValue->u.str->strData;
			}
			
			// in the deaactivate routing
			if (myStruct->mString != NULL) {
				delete myStruct->mString;
			}



			LoadJSON(ip, inActorInfo, inNewValue->u.str->strData);
		}
		break;

	case kInputItemCount:
		{
			// delta is how many more or less inputs we want, as compared to the current count
			int delta = inNewValue->u.ivalue - variableInputCount;
			
			Value valueInit;
			valueInit.type = kFloat;
			valueInit.u.fvalue = 0.0f;
			
			if (delta > 0) {
			
				while (delta-- > 0) {
				
					char propertyName[256];
					sprintf(propertyName, "select %d", (int)(variableInputCount + 1));
					
					OSType inputID = CreatePropertyID(ip, "i", variableInputCount + 1);
					
					
					Value valueStr = { 0, { 0 } };
					AllocateValueString_(ip, "", &valueStr);
					
					Value valueInit;
					valueInit.type = kString;
					valueInit.u.str = NULL;
					Value valueMin;
					valueMin.type = kString;
					valueMin.u.fvalue = NULL;
					Value valueMax;
					valueMax.type = kString;
					valueMax.u.fvalue = NULL;
					
					err = AddPropertyCustomDataTypeAfterIndex_(ip, inActorInfo,
						kInputProperty,
						inputID,					// the property ID
						'i001',						// the input to which we will conform
						propertyName,
						0,							// custom data type -- n/a
						kDisplayFormatText,
						kDisplayFormatText,
						1,
						&valueStr,
						&valueStr,
						&valueStr,
						kFixedInputPropertyCountBefore + variableInputCount);
					
					PluginAssert_(ip, err == noErr);

					sprintf(propertyName, "element %d", (int)(variableInputCount + 1));
					OSType outputID = CreatePropertyID(ip, "o", variableInputCount + 1);
					#if MUTABLE_ITEMS
						err = AddProperty_(ip, inActorInfo, 
							kOutputProperty,
							outputID,						// the property ID
							'o001',							// the output to which we will conform
							propertyName,
							kDisplayFormatText | kDisplayFormatMutable,
							kDisplayFormatText,
							kIndMutableGroup,
							&valueStr,
							&valueStr,
							&valueStr);
					#else
						err = AddProperty_(ip, inActorInfo, 
							kOutputProperty,
							outputID,						// the property ID
							'o001',							// the output to which we will conform
							propertyName,
							kDisplayFormatText,
							kDisplayFormatText,
							0,
							&valueStr,
							&valueStr,
							&valueStr);
					#endif
					PluginAssert_(ip, err == noErr);
					
					ReleaseValueString_(ip, &valueStr);
					
					variableInputCount++;
				}
				
			} else if (delta < 0) {
				
				while (delta < 0) {
					err = RemoveProperty_(ip, inActorInfo, kInputProperty, variableInputCount + kFixedInputPropertyCountBefore);
					PluginAssert_(ip, err == noErr);
					err = RemoveProperty_(ip, inActorInfo, kOutputProperty, variableInputCount + kFixedOutputPropertyCount);
					PluginAssert_(ip, err == noErr);
					delta++;
					variableInputCount--;
				}
			}
			
			// UpdateOutputTypes(ip, inActorInfo);
		}
		break;
		
		case kInput1Base:
		{
			if (item >= 0 && item < kMaxOutputNeedsUpdateFlags) {
				info->mOutputNeedsUpdate[item] = true;
			}
		}
		break;
	}

	if (!inInitializing) {
		for (PropertyIndex i=0; i<variableInputCount; i++) {
			if (i < kMaxOutputNeedsUpdateFlags && info->mOutputNeedsUpdate[i]) {
				ValuePtr v = GetInputPropertyValue_(ip, inActorInfo, kInput1Base + i);
				ProcessInputValue(ip, inActorInfo, kFirstVariableOutputProperty+i, v);
				info->mOutputNeedsUpdate[i] = false;
			}
		}
	}
}

// ---------------------------------------------------------------------------------
//		 PropertyValueToString
// ---------------------------------------------------------------------------------
//	Handles incoming video frame changed messages

const char* sErrorFlag[] = {
	"no",
	"yes"
};

static Boolean
PropertyValueToString(
	IsadoraParameters*	ip,
	ActorInfo*			/* inActorInfo */,
	PropertyType		inPropertyType,
	PropertyIndex		inPropertyIndex1,
	ValuePtr			inValue,
	char*				outString)
{
	Boolean result = false;
	
	if (inPropertyType == kOutputProperty && inPropertyIndex1 == kOutputError) {
		switch (inValue->u.ivalue) {
		case 0:
		case 1:
			strcpy(outString, sErrorFlag[inValue->u.ivalue]);
			break;
		default:
			PluginAssert_(ip, false);
			break;
		}
		result = true;
	}
	
	return result;
}

// ---------------------------------------------------------------------------------
//		 PropertyStringToValue
// ---------------------------------------------------------------------------------

static Boolean
PropertyStringToValue(
	IsadoraParameters*	ip,
	ActorInfo*			/* inActorInfo */,
	PropertyType		inPropertyType,
	PropertyIndex		inPropertyIndex1,
	const char*			inString,
	ValuePtr			outValue)
{
	Boolean encodeSuccessful = false;
	
	if (inPropertyType == kOutputProperty && inPropertyIndex1 == kOutputError) {
		if (strlen(inString) == 1 && inString[0] >= '0' && inString[0] <= '1') {
			outValue->type = kInteger;
			outValue->u.ivalue = (SInt32) inString[0] - '0';
			encodeSuccessful = true;
		} else {
			SInt32 matchIndex = LookupPartialStringInList_(ip, 2, sErrorFlag, inString);
			if (matchIndex >= 0) {
				outValue->type = kInteger;
				outValue->u.ivalue = matchIndex;
				encodeSuccessful = true;
			}
		}
	}
	
	return encodeSuccessful;
}

// ---------------------------------------------------------------------------------
//		 GetActorDefinedArea
// ---------------------------------------------------------------------------------

static ActorPictInfo	gPictInfo = { false, nil, nil, 0, 0 };

static Boolean
GetActorDefinedArea(
	IsadoraParameters*			ip,			
	ActorInfo*					inActorInfo,
	SInt16*						outTopAreaWidth,
	SInt16*						outTopAreaMinHeight,
	SInt16*						outBotAreaHeight,
	SInt16*						outBotAreaMinWidth)
{
	if (gPictInfo.mInitialized == false) {
		PrepareActorDefinedAreaPict_(ip, inActorInfo, 0, &gPictInfo);
	}
	
	*outTopAreaWidth = gPictInfo.mWidth;
	*outTopAreaMinHeight = gPictInfo.mHeight;
	*outBotAreaHeight = 0;
	*outBotAreaMinWidth = 0;

	
	return true;
}

// ---------------------------------------------------------------------------------
//		 DrawActorDefinedArea
// ---------------------------------------------------------------------------------

static void
DrawActorDefinedArea(
	IsadoraParameters*			ip,
	ActorInfo*					inActorInfo,
	void*						/* inDrawingContext */,
	ActorDefinedAreaPart		inActorDefinedAreaPart,
	ActorAreaDrawFlagsT			/* inAreaDrawFlags */,
	Rect*						inADAArea,
	Rect*						/* inUpdateArea */,
	Boolean						inSelected)
{
	if (inActorDefinedAreaPart == kActorDefinedAreaTop && gPictInfo.mInitialized) {
		DrawActorDefinedAreaPict_(ip, inActorInfo, inSelected, inADAArea, &gPictInfo);
	}
}
	
// ---------------------------------------------------------------------------------
//		 HandlePropertyChangeType
// ---------------------------------------------------------------------------------

static void
HandlePropertyChangeType(
	IsadoraParameters*					ip,
	ActorInfo*							inActorInfo,
	const PropertyDefinedValueSource*	/* inPropDefValueSrc */,
	PropertyType						inPropertyType,
	PropertyIndex						inPropertyIndex1,
	PropertyDispFormat					/* inFormats */,
	PropertyDispFormat					/* inCurFormat */,
	ValuePtr							/* inMinimum */,
	ValuePtr							/* inMaximum */,
	ValuePtr							/* inScaleMin */,
	ValuePtr							/* inScaleMax */,
	ValuePtr							/* inValue */)
{
	if (gInsideHandlePropertyChangeType == false) {
	
		gInsideHandlePropertyChangeType = true;
		
		if (inPropertyType == kOutputProperty && inPropertyIndex1 >= kFirstVariableOutputProperty) {
			// log_info("Changed Output %d - New Type = %d\n", (int) inPropertyIndex1, (int) inValue->type);
			// UpdateOutputTypes(ip, inActorInfo);
		}
		
		gInsideHandlePropertyChangeType = false;
	}
}

// ---------------------------------------------------------------------------------
//		 GetActorInfo
// ---------------------------------------------------------------------------------
//	This is function is called by to get the actor's class and ID, and to get
//	pointers to the all of the plugin functions declared locally.
//
//	All members of the ActorInfo struct pointed to by outActorParams have been
//	set to 0 on entry. You only need set functions defined by your plugin
//	
EXPORT_ void
GetActorInfo(
	void*				/* inParam */,
	ActorInfo*			outActorParams)
{
	// REQUIRED information
	outActorParams->mActorName							= kActorName;
	outActorParams->mClass								= kActorClass;
	outActorParams->mID									= kActorID;
	outActorParams->mCompatibleWithVersion				= kActorCompatilibilityVersion_v1;
	outActorParams->mActorFlags							= kActorFlags_Plugin_CheckForUpdates;
	
	// REQUIRED functions
	outActorParams->mGetActorParameterStringProc		= GetParameterString;
	outActorParams->mGetActorHelpStringProc				= GetHelpString;
	outActorParams->mCreateActorProc					= CreateActor;
	outActorParams->mDisposeActorProc					= DisposeActor;
	outActorParams->mActivateActorProc					= ActivateActor;
	outActorParams->mHandlePropertyChangeValueProc		= HandlePropertyChangeValue;
	
	// OPTIONAL FUNCTIONS
	outActorParams->mHandlePropertyChangeTypeProc		= HandlePropertyChangeType;
	outActorParams->mHandlePropertyConnectProc			= NULL;
	outActorParams->mPropertyValueToStringProc			= PropertyValueToString;
	outActorParams->mPropertyStringToValueProc			= PropertyStringToValue;
	outActorParams->mGetActorDefinedAreaProc			= GetActorDefinedArea;
	outActorParams->mDrawActorDefinedAreaProc			= DrawActorDefinedArea;
	outActorParams->mMouseTrackInActorDefinedAreaProc	= NULL;

	outActorParams->mSetActorPrivateDataProc			= NULL;
	outActorParams->mGetActorPrivateDataLengthProc		= NULL;
	outActorParams->mGetActorPrivateDataProc			= NULL;
}
